from typing import Any, Literal, TypedDict
from .common import BaseOperator
from ..models import SmartCard, StatusCodes
from ..util import StatusInteractive, ExecutionError


class FetchedKeyResult(TypedDict):
    id: str
    first_name: str
    last_name: str
    extras: str


class CardOperator(BaseOperator):

    def __init__(self, gpg: Any, interactive: StatusInteractive) -> None:
        super().__init__(gpg)
        self.interactive = interactive

    def debug(self):
        for i in self.interactive.readlines(yield_empty=False):
            print(i)

    @property
    def active(self) -> SmartCard | None:
        """Gets information about the current card.

        Returns:
            SmartCard | None: Card data, or None if no card is present.
        """
        result = self.session.run("gpg --with-colons --card-status")
        if result.code == 0:
            return SmartCard.from_status(result.output)
        else:
            return None

    def reset(self) -> SmartCard:
        """Resets the active card to factory settings.

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Reset card instance
        """
        self.interactive.writelines("factory-reset")
        entered = False
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                lines.append(line.content)

                if (
                    cmd == StatusCodes.GET_BOOL
                    and arg == "cardedit.factory-reset.proceed"
                ):
                    self.interactive.writelines("y")

                elif cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.factory-reset.really":
                        self.interactive.writelines("yes")
                        success = True
                    elif arg == "cardedit.prompt":
                        if entered:
                            if success:
                                return self.active
                            else:
                                raise ExecutionError("\n".join(lines))
                        else:
                            entered = True

    def set_name(
        self, first_name: str, last_name: str, admin_pin: str = "12345678"
    ) -> SmartCard:
        """Sets the name of the cardholder

        Args:
            first_name (str): First Name
            last_name (str): Last Name
            admin_pin (str, optional): Admin PIN of card. Defaults to "12345678".

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated SmartCard
        """
        self.interactive.writelines("name")
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "keygen.smartcard.surname":
                        self.interactive.writelines(last_name)
                    elif arg == "keygen.smartcard.givenname":
                        self.interactive.writelines(first_name)
                    elif arg == "cardedit.prompt":
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))
                    else:
                        self.interactive.writelines("")
                elif cmd == StatusCodes.GET_HIDDEN:
                    self.interactive.writelines(admin_pin)
                elif cmd == StatusCodes.SC_OP_SUCCESS:
                    success = True

    def set_key_url(self, url: str | None, admin_pin: str = "12345678") -> SmartCard:
        """Sets the URL of the current card's public key.

        Args:
            url (str | None): URL, or None to unset
            admin_pin (str, optional): Card admin PIN. Defaults to "12345678".

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card instance
        """
        self.interactive.writelines("url")
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.change_url":
                        self.interactive.writelines(url if url else "UNSET")
                    else:
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))
                elif cmd == StatusCodes.GET_HIDDEN:
                    self.interactive.writelines(admin_pin)
                elif cmd == StatusCodes.SC_OP_SUCCESS:
                    success = True

    def get_key_from_url(self) -> list[FetchedKeyResult] | None:
        """Gets the key information from the card's URL, returning None if not present.

        Returns:
            list[FetchedKeyResult] | None: List of results, or None if not present.
        """
        if not self.active.public_key_url or self.active.public_key_url == "UNSET":
            return None

        self.interactive.writelines("fetch")
        results: list[FetchedKeyResult] = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status and line.code == StatusCodes.IMPORTED:
                results.append(
                    {
                        "id": line.arguments[0],
                        "first_name": (
                            line.arguments[1] if len(line.arguments) > 1 else ""
                        ),
                        "last_name": (
                            line.arguments[2] if len(line.arguments) > 2 else ""
                        ),
                        "extras": (
                            " ".join(line.arguments[3:])
                            if len(line.arguments) > 3
                            else ""
                        ),
                    }
                )
            elif line.code == StatusCodes.GET_LINE:
                return results

    def set_login(self, login: str, admin_pin: str = "12345678") -> SmartCard:
        self.interactive.writelines("login")
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.change_login":
                        self.interactive.writelines(login)
                    else:
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))
                elif cmd == StatusCodes.GET_HIDDEN:
                    self.interactive.writelines(admin_pin)
                elif cmd == StatusCodes.SC_OP_SUCCESS:
                    success = True

    def set_language(self, *languages: str, admin_pin: str = "12345678") -> SmartCard:
        """Set language preferences

        Args:
            *languages (str): 1-4 ISO-639 language codes.
            admin_pin (str, optional): Admin PIN. Defaults to "12345678".

        Raises:
            ValueError: If specified arguments are invalid.
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        if len(languages) == 0:
            raise ValueError("At least one language must be specified.")
        if len(languages) > 4:
            raise ValueError("At most 4 languages must be specified.")
        if any([len(i) != 2 for i in languages]):
            raise ValueError(
                "All language entries must be 2-byte ISO-639 language codes."
            )

        self.interactive.writelines("lang")
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.change_lang":
                        self.interactive.writelines("".join(languages))
                    else:
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))
                elif cmd == StatusCodes.GET_HIDDEN:
                    self.interactive.writelines(admin_pin)
                elif cmd == StatusCodes.SC_OP_SUCCESS:
                    success = True

    def set_salutation(
        self,
        salutation: Literal["male", "female"] | None = None,
        admin_pin: str = "12345678",
    ) -> SmartCard:
        """Sets the cardholder's salutation preferences.

        Args:
            salutation ('male' | 'female' | None, optional): Saltuation mode, or unset if None. Defaults to None.
            admin_pin (str, optional): Admin PIN. Defaults to "12345678".

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated smartcard
        """
        self.interactive.writelines("salutation")
        success = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.change_sex":
                        self.interactive.writelines(
                            ("M" if salutation == "male" else "F")
                            if salutation
                            else " "
                        )
                    else:
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))
                elif cmd == StatusCodes.GET_HIDDEN:
                    self.interactive.writelines(admin_pin)
                elif cmd == StatusCodes.SC_OP_SUCCESS:
                    success = True

    def set_forced_sig(self, value: bool, admin_pin: str = "12345678") -> SmartCard:
        """Set forced signature mode

        Args:
            value (bool): What value to set to
            admin_pin (str, optional): Admin PIN. Defaults to "12345678".

        Returns:
            SmartCard: Updated smartcard
        """
        current = self.active
        if current.forced_signature_pin == value:
            return current

        self.interactive.writelines("forcesig")
        line = self.interactive.wait_for_status(
            StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN
        )[-1]
        if line.code == StatusCodes.GET_HIDDEN:
            self.interactive.writelines(admin_pin)
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
        return self.active

    def generate_key(
        self,
        real_name: str,
        email: str | None = None,
        comment: str | None = None,
        expires: str | None = None,
        backup: bool = False,
        card_pin: str = "123456",
        admin_pin: str = "12345678",
        key_passphrase: str | None = None,
        force: bool = False,
    ) -> SmartCard:
        """Generates a key on the smartcard, optionally backing up to the local machine.

        Args:
            real_name (str): UID name
            email (str | None, optional): UID email. Defaults to None.
            comment (str | None, optional): UID comment. Defaults to None.
            expires (str | None, optional): Expiration time (None for no expiration, `n` for n days, `nW` for n weeks, etc). Defaults to None.
            backup (bool, optional): Whether to backup to the local machine. Defaults to False.
            card_pin (str, optional): Card PIN. Defaults to "123456".
            admin_pin (str, optional): Card Admin PIN. Defaults to "12345678".
            key_passphrase (str | None, optional): The passphrase for the local backup key. Defaults to None.
            force (bool, optional): Whether to replace existing keys. Defaults to False.

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("generate")
        success = False
        pinentered = False
        admentered = False
        lines = []
        for line in self.interactive.readlines(yield_empty=False):
            if line.is_status:
                lines.append(line.content)
                cmd = line.code
                arg = line.arguments[0] if len(line.arguments) > 0 else None
                if cmd == StatusCodes.GET_LINE:
                    if arg == "cardedit.genkeys.backup_enc":
                        self.interactive.writelines("y" if backup else "n")
                    elif arg == "keygen.valid":
                        self.interactive.writelines(expires if expires else "0")
                    elif arg == "keygen.name":
                        self.interactive.writelines(real_name)
                    elif arg == "keygen.email":
                        self.interactive.writelines(email if email else "")
                    elif arg == "keygen.comment":
                        self.interactive.writelines(comment if comment else "")
                    elif arg == "cardedit.prompt":
                        if success:
                            return self.active
                        else:
                            raise ExecutionError("\n".join(lines))

                elif cmd == StatusCodes.GET_HIDDEN:
                    if arg == "passphrase.enter":
                        if pinentered:
                            if admentered:
                                if backup:
                                    self.interactive.writelines(
                                        key_passphrase if key_passphrase else ""
                                    )

                            else:
                                self.interactive.writelines(admin_pin)
                                admentered = True
                        else:
                            self.interactive.writelines(card_pin)
                            pinentered = True

                elif (
                    cmd == StatusCodes.GET_BOOL
                    and arg == "cardedit.genkeys.replace_keys"
                ):
                    self.interactive.writelines("y" if force else "n")

                elif cmd == StatusCodes.KEY_CREATED:
                    success = True

    def change_pin(self, current_pin: str, new_pin: str) -> SmartCard:
        """Change the user PIN of the key

        Args:
            current_pin (str): Current user PIN
            new_pin (str): New user PIN

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("passwd")
        line = self.interactive.wait_for_status(StatusCodes.GET_LINE)[-1]
        if line.arguments[0] == "cardutil.change_pin.menu":
            self.interactive.writelines("1", current_pin, new_pin, new_pin)
        else:
            raise ExecutionError("Failed to change PIN")

        line = self.interactive.wait_for_status(
            StatusCodes.SC_OP_SUCCESS, StatusCodes.SC_OP_FAILURE
        )[-1]
        self.interactive.wait_for_status(StatusCodes.GET_LINE)
        if line.code == StatusCodes.SC_OP_SUCCESS:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            return self.active
        else:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            raise ExecutionError("Failed to change PIN")

    def unblock_pin_as_admin(self, admin_pin: str, new_pin: str) -> SmartCard:
        """Unblock a user PIN with the admin PIN

        Args:
            admin_pin (str): Admin PIN
            new_pin (str): New user PIN

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("passwd")
        line = self.interactive.wait_for_status(StatusCodes.GET_LINE)[-1]
        if line.arguments[0] == "cardutil.change_pin.menu":
            self.interactive.writelines("2", admin_pin, new_pin, new_pin)
        else:
            raise ExecutionError("Failed to unblock PIN")

        line = self.interactive.wait_for_status(
            StatusCodes.SC_OP_SUCCESS, StatusCodes.SC_OP_FAILURE
        )[-1]
        self.interactive.wait_for_status(StatusCodes.GET_LINE)
        if line.code == StatusCodes.SC_OP_SUCCESS:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            return self.active
        else:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            raise ExecutionError("Failed to unblock PIN")

    def unblock_pin(self, reset_code: str, new_pin: str) -> SmartCard:
        """Unblock the user PIN with a reset code.

        Args:
            reset_code (str): Reset code
            new_pin (str): New user PIN

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("unblock")
        line = self.interactive.wait_for_status(
            StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN
        )[-1]
        if line.code == StatusCodes.GET_LINE:
            raise ExecutionError("Reset code is unavailable or otherwise inaccessible.")

        self.interactive.writelines(reset_code, new_pin, new_pin)
        line = self.interactive.wait_for_status(
            StatusCodes.SC_OP_SUCCESS, StatusCodes.SC_OP_FAILURE
        )[-1]
        self.interactive.wait_for_status(StatusCodes.GET_LINE)
        if line.code == StatusCodes.SC_OP_SUCCESS:
            return self.active
        else:
            raise ExecutionError("Failed to unblock PIN")

    def change_admin_pin(self, current_pin: str, new_pin: str) -> SmartCard:
        """Change the admin PIN

        Args:
            current_pin (str): Current admin PIN
            new_pin (str): New admin PIN

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("passwd")
        line = self.interactive.wait_for_status(StatusCodes.GET_LINE)[-1]
        if line.arguments[0] == "cardutil.change_pin.menu":
            self.interactive.writelines("3", current_pin, new_pin, new_pin)
        else:
            raise ExecutionError("Failed to change Admin PIN")

        line = self.interactive.wait_for_status(
            StatusCodes.SC_OP_SUCCESS, StatusCodes.SC_OP_FAILURE
        )[-1]
        self.interactive.wait_for_status(StatusCodes.GET_LINE)
        if line.code == StatusCodes.SC_OP_SUCCESS:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            return self.active
        else:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            raise ExecutionError("Failed to change Admin PIN")

    def change_reset_code(self, admin_pin: str, reset_code: str) -> SmartCard:
        """Changes the card's reset code

        Args:
            admin_pin (str): Admin PIN
            reset_code (str): New reset code

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        self.interactive.writelines("passwd")
        line = self.interactive.wait_for_status(StatusCodes.GET_LINE)[-1]
        if line.arguments[0] == "cardutil.change_pin.menu":
            self.interactive.writelines("4", admin_pin, reset_code, reset_code)
        else:
            raise ExecutionError("Failed to change reset code")

        line = self.interactive.wait_for_status(
            StatusCodes.SC_OP_SUCCESS, StatusCodes.SC_OP_FAILURE
        )[-1]
        self.interactive.wait_for_status(StatusCodes.GET_LINE)
        if line.code == StatusCodes.SC_OP_SUCCESS:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            return self.active
        else:
            self.interactive.writelines("Q")
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            raise ExecutionError("Failed to change reset code")

    def set_usage_info(
        self, type: Literal["sign", "decrypt", "auth"], value: bool, admin_pin: str
    ) -> SmartCard:
        """Sets card UIF data

        Args:
            type (sign | decrypt | auth): Which UIF flag to set
            value (bool): What value to set it to
            admin_pin (str): Card admin PIN

        Raises:
            ExecutionError: If operation fails

        Returns:
            SmartCard: Updated card
        """
        indexMap = {"sign": 1, "decrypt": 2, "auth": 3}
        self.interactive.writelines(f"uif {indexMap[type]} {'on' if value else 'off'}")
        line = self.interactive.wait_for_status(
            StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN
        )[-1]
        if line.code == StatusCodes.GET_LINE:
            raise ExecutionError("Input is invalid.")

        self.interactive.writelines(admin_pin)
        line = self.interactive.wait_for_status(
            StatusCodes.GET_LINE, StatusCodes.SC_OP_FAILURE
        )[-1]
        if line.code == StatusCodes.GET_LINE:
            return self.active
        else:
            self.interactive.wait_for_status(StatusCodes.GET_LINE)
            raise ExecutionError("Incorrect PIN")



from collections.abc import Generator
import subprocess
from tempfile import NamedTemporaryFile
from typing import Any

from pydantic import BaseModel, computed_field
from .process import ProcessSession


class StatusLine(BaseModel):
    content: str
    code: str | None = None
    arguments: list[str] | None = None

    @computed_field
    @property
    def is_status(self) -> bool:
        return self.code != None

    @classmethod
    def from_line(cls, line: bytes) -> "StatusLine":
        decoded = line.decode().rstrip("\n")
        if decoded.startswith("[GNUPG:]"):
            return StatusLine(
                content=decoded,
                code=decoded.split(" ")[1].strip(),
                arguments=decoded.split(" ")[2:],
            )
        else:
            return StatusLine(content=decoded)


class Interactive:
    def __init__(
        self,
        session: ProcessSession,
        command: str | list[str],
        shell: bool | None = None,
        environment: dict[str, str] | None = None,
        working_directory: str | None = None,
    ):
        self.session = session
        self.options = self.session.make_kwargs(
            shell=shell, env=environment, cwd=working_directory
        )
        self.parsed_command = self.session.parse_cmd(
            command, shell=bool(self.options.get("shell", False))
        )

        self.output_file = None
        self.output_handle = None
        self.process = None
        self.code = None

    def __enter__(self) -> "Interactive":
        self.output_file = NamedTemporaryFile()
        self.output_handle = open(self.output_file.name, "rb")
        self.process = subprocess.Popen(
            self.parsed_command,
            stdin=subprocess.PIPE,
            stdout=self.output_file,
            stderr=self.output_file,
            **self.options,
        )
        return self

    def __exit__(self, *args, **kwargs):
        if self.process.poll() == None:
            self.process.terminate()
            try:
                self.process.wait(timeout=0)
            except:
                pass
        self.code = self.process.poll()
        self.output_handle.close()
        self.output_file.close()
        del self.process

    def seek(self, position: int = 0):
        self.output_handle.seek(position)

    def read(self, amount: int = -1) -> bytes | None:
        return self.output_handle.read(amount)

    def readline(self) -> bytes | None:
        line = self.output_handle.readline()
        if len(line) == 0:
            return None
        return line

    def readlines(self, yield_empty: bool = True) -> Generator[bytes | None, Any, Any]:
        while True:
            try:
                line = self.readline()
                if line != None or yield_empty:
                    yield line
            except:
                break

    def write(self, content: bytes):
        self.process.stdin.write(content)
        self.process.stdin.flush()

    def writelines(self, *lines: bytes | str):
        concatenated = (
            b"\n".join(
                [
                    i.encode().rstrip(b"\n") if type(i) == str else i.rstrip(b"\n")
                    for i in lines
                ]
            )
            + b"\n"
        )
        self.write(concatenated)


class StatusInteractive(Interactive):
    def readline(self) -> StatusLine | None:
        line = super().readline()
        return StatusLine.from_line(line) if line else None

    def __enter__(self) -> "StatusInteractive":
        return super().__enter__()

    def readlines(
        self, yield_empty: bool = True
    ) -> Generator[StatusLine | None, Any, Any]:
        while True:
            try:
                line = self.readline()
                if line != None or yield_empty:
                    yield line
            except:
                break

    def wait_for_status(self, *code: str):
        lines: list[StatusLine] = []
        for line in self.readlines():
            if line:
                lines.append(line)
                if line.is_status and (len(code) == 0 or line.code in code):
                    return lines



import datetime
from enum import IntEnum, StrEnum
from pydantic import BaseModel, computed_field


class InfoRecord(StrEnum):
    """Describes the type of an individual record."""
    PUBLIC_KEY = "pub"
    X509_CERTIFICATE = "crt"
    X509_CERTIFICATE_WITH_SECRET = "crs"
    SUBKEY = "sub"
    SECRET_KEY = "sec"
    SECRET_SUBKEY = "ssb"
    USER_ID = "uid"
    USER_ATTRIBUTE = "uat"
    SIGNATURE = "sig"
    REVOCATION_SIGNATURE = "rev"
    REVOCATION_SIGNATURE_STANDALONE = "rvs"
    FINGERPRINT = "fpr"
    SHA256_FINGERPRINT = "fp2"
    PUBLIC_KEY_DATA = "pkd"
    KEYGRIP = "grp"
    REVOCATION_KEY = "rvk"
    TOFU_DATA = "tfs"
    TRUST_INFO = "tru"
    SIGNATURE_SUBPACKET = "spk"
    CONFIG_DATA = "cfg"


class FieldValidity(StrEnum):
    """Describes the validity of a specific record, for instance that of a key or UID"""
    UNKNOWN = "o"
    INVALID = "i"
    DISABLED = "d"
    REVOKED = "r"
    EXPIRED = "e"
    UNKNOWN_VALIDITY = "-"
    UNDEFINED_VALIDITY = "q"
    NOT_VALID = "n"
    MARGINAL_VALID = "m"
    FULLY_VALID = "f"
    ULTIMATELY_VALID = "u"
    WELL_KNOWN = "w"
    SPECIAL = "s"


class SignatureValidity(StrEnum):
    """Describes the validity of an individual signature."""
    GOOD = "!"
    BAD = "-"
    NO_PUBLIC_KEY = "?"
    UNKNOWN_ERROR = "%"


class KeyCapability(StrEnum):
    """Describes a capability of a key."""
    ENCRYPT = "e"
    SIGN = "s"
    CERTIFY = "c"
    AUTHENTICATION = "a"
    RESTRICTED_ENCRYPTION = "r"
    TIMESTAMPING = "t"
    GROUP_KEY = "g"
    UNKNOWN = "?"
    DISABLED = "d"


class StaleTrustReason(StrEnum):
    OLD = "o"
    DIFFERENT_MODEL = "t"


class TrustModel(IntEnum):
    CLASSIC = 0
    PGP = 1


class InfoLine(BaseModel):
    record_type: InfoRecord
    field_array: list[str | None]

    def field(self, field: int) -> str | None:
        """Get field value based on indices from https://github.com/gpg/gnupg/blob/master/doc/DETAILS

        Args:
            field (int): Field number (1-21)

        Raises:
            KeyError: If field is unknown

        Returns:
            str | None: Field value or None if empty.
        """
        if field < 1 or field > 21:
            raise KeyError("Unknown field number.")

        if field == 1:
            return self.record_type

        try:
            return self.field_array[field - 2]
        except:
            return None

    @classmethod
    def from_line(cls, line: str) -> "InfoLine":
        parts = line.split(":")
        return cls(
            record_type=parts[0],
            field_array=[i if len(i) > 0 else None for i in parts[1:]],
        )

    @property
    def fields(self) -> list[str | None]:
        return [self.field(i) for i in range(2, 22)]


# pub, sub
class KeyInfo(InfoLine):
    """Raw information about a single Key record"""

    @computed_field
    @property
    def validity(self) -> FieldValidity | None:
        """Key validity

        Returns:
            FieldValidity | None: Key validity, if available
        """
        return FieldValidity(self.field(2)) if self.field(2) else None

    @computed_field
    @property
    def length(self) -> int:
        """Key length

        Returns:
            int: Key length, or 0 if unavailable
        """
        return int(self.field(3))

    @computed_field
    @property
    def algorithm(self) -> int:
        """The algorithm ID used for this key

        Returns:
            int: Algorithm ID
        """
        return int(self.field(4))

    @computed_field
    @property
    def key_id(self) -> str:
        """The key's ID

        Returns:
            str: Key ID
        """
        return self.field(5)

    @computed_field
    @property
    def creation_date(self) -> datetime.datetime | None:
        """The key's creation date

        Returns:
            datetime.datetime | None: Creation date, if available
        """
        value = self.field(6)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def expiration_date(self) -> datetime.datetime | None:
        """The key's expiration date

        Returns:
            datetime.datetime | None: Expiration date, if available
        """
        value = self.field(7)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def owner_trust(self) -> str | None:
        """The owner trust value

        Returns:
            str | None: Owner trust, if available
        """
        return self.field(9)

    @computed_field
    @property
    def capabilities(self) -> list[KeyCapability]:
        """Key capabilities

        Returns:
            list[KeyCapability]: A list of all capabilities associated with this key
        """
        value = self.field(12)
        if value == None:
            return []

        return [KeyCapability(i) for i in value if i.lower() == i]

    @computed_field
    @property
    def overall_capabilities(self) -> list[KeyCapability]:
        """Overall key capabilities

        Returns:
            list[KeyCapability]: A list of all capabilities associated with this key and its subkeys
        """
        value = self.field(12)
        if value == None:
            return []

        return [KeyCapability(i.lower()) for i in value if i.lower() != i]

    @computed_field
    @property
    def curve_name(self) -> str | None:
        """The name of this key's ECC curve

        Returns:
            str | None: Curve name, if available
        """
        return self.field(17)


# fpr, fp2
class FingerprintInfo(InfoLine):
    @computed_field
    @property
    def fingerprint(self) -> str:
        return self.field(10)


# grp
class KeygripInfo(InfoLine):

    @computed_field
    @property
    def keygrip(self) -> str:
        return self.field(10)


# sec, ssb
class SecretKeyInfo(KeyInfo):
    """An extension to KeyInfo, adding S/N info"""

    @computed_field
    @property
    def serial_number(self) -> str | None:
        """The serial number of this key

        Returns:
            str | None: S/N, if available
        """
        return self.field(15)


# uid
class UserIDInfo(InfoLine):
    """Info pertaining to a single UID"""

    @computed_field
    @property
    def validity(self) -> FieldValidity | None:
        """This UID's validity

        Returns:
            FieldValidity | None: UID validity, if available
        """
        return FieldValidity(self.field(2)) if self.field(2) else None

    @computed_field
    @property
    def creation_date(self) -> datetime.datetime | None:
        """UID creation date

        Returns:
            datetime.datetime | None: Creation date, if available
        """
        value = self.field(6)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def expiration_date(self) -> datetime.datetime | None:
        """UID expiration date

        Returns:
            datetime.datetime | None: Expiration date, if available
        """
        value = self.field(7)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def uid_hash(self) -> str | None:
        """The UID hash

        Returns:
            str | None: UID hash, if available
        """
        return self.field(8)

    @computed_field
    @property
    def uid(self) -> str:
        """Full UID string

        Returns:
            str: The UID string
        """
        return self.field(10)


# sig, rev
class SignatureInfo(InfoLine):
    """Info about a single signature (revocation & non-revocation)"""

    @computed_field
    @property
    def is_revocation(self) -> bool:
        """Whether this signature is a revocation signature

        Returns:
            bool: True if revocation, false otherwise
        """
        return self.record_type == InfoRecord.REVOCATION_SIGNATURE

    @computed_field
    @property
    def validity(self) -> SignatureValidity | None:
        """The validity of this signature

        Returns:
            SignatureValidity | None: The signature's validity, if available
        """
        value = self.field(2)
        if value and len(value) > 0:
            return SignatureValidity(value[0])
        else:
            return None

    @computed_field
    @property
    def algorithm(self) -> int:
        """The signature's algorithm

        Returns:
            int: Algorithm ID
        """
        return int(self.field(4))

    @computed_field
    @property
    def key_id(self) -> str:
        """The signer's key ID

        Returns:
            str: Key ID
        """
        return self.field(5)

    @computed_field
    @property
    def creation_date(self) -> datetime.datetime | None:
        """Signature creation date

        Returns:
            datetime.datetime | None: Creation date, if available
        """
        value = self.field(6)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def expiration_date(self) -> datetime.datetime | None:
        """The signature's expiration date

        Returns:
            datetime.datetime | None: Expiration date, if available
        """
        value = self.field(7)
        if value:
            if "T" in value:
                return datetime.datetime.fromisoformat(value)
            else:
                return datetime.datetime.fromtimestamp(float(value))
        return None

    @computed_field
    @property
    def uid(self) -> str:
        """The UID of the signature

        Returns:
            str: Signature UID
        """
        return self.field(10)

    @computed_field
    @property
    def signature_class(self) -> str:
        """The signature class

        Returns:
            str: Signature class info
        """
        return self.field(11)

    @computed_field
    @property
    def signer_fingerprint(self) -> str:
        """The fingerprint of the signer's key

        Returns:
            str: Key fingerprint
        """
        return self.field(13)


# tru
class TrustInfo(InfoLine):

    @computed_field
    @property
    def staleness(self) -> None | StaleTrustReason:
        return StaleTrustReason(self.field(2)) if self.field(2) else None

    @computed_field
    @property
    def trust_model(self) -> TrustModel:
        return TrustModel(int(self.field(3)))

    @computed_field
    @property
    def creation_date(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(float(self.field(4)))

    @computed_field
    @property
    def expiration_date(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(float(self.field(5)))

    @computed_field
    @property
    def marginals_needed(self) -> int:
        return int(self.field(6))

    @computed_field
    @property
    def completes_needed(self) -> int:
        return int(self.field(7))

    @computed_field
    @property
    def max_cert_depth(self) -> int:
        return int(self.field(8))


def parse_infoline(line: str) -> InfoLine:
    """Parses a raw infoline into either a generic InfoLine or one of the specific types

    Args:
        line (str): Line to parse

    Returns:
        InfoLine: InfoLine or a subclass
    """
    initial_parse = InfoLine.from_line(line)
    match initial_parse.record_type:
        case InfoRecord.PUBLIC_KEY:
            return KeyInfo.from_line(line)

        case InfoRecord.SUBKEY:
            return KeyInfo.from_line(line)

        case InfoRecord.SECRET_KEY:
            return SecretKeyInfo.from_line(line)

        case InfoRecord.SECRET_SUBKEY:
            return SecretKeyInfo.from_line(line)

        case InfoRecord.FINGERPRINT:
            return FingerprintInfo.from_line(line)

        case InfoRecord.SHA256_FINGERPRINT:
            return FingerprintInfo.from_line(line)

        case InfoRecord.USER_ID:
            return UserIDInfo.from_line(line)

        case InfoRecord.SIGNATURE:
            return SignatureInfo.from_line(line)

        case InfoRecord.REVOCATION_SIGNATURE:
            return SignatureInfo.from_line(line)

        case InfoRecord.TRUST_INFO:
            return TrustInfo.from_line(line)

        case InfoRecord.KEYGRIP:
            return KeygripInfo.from_line(line)

        case _:
            return initial_parse



from enum import StrEnum


class StatusCodes(StrEnum):
    NEWSIG = "NEWSIG"
    GOODSIG = "GOODSIG"
    EXPSIG = "EXPSIG"
    EXPKEYSIG = "EXPKEYSIG"
    REVKEYSIG = "REVKEYSIG"
    BADSIG = "BADSIG"
    ERRSIG = "ERRSIG"
    VALIDSIG = "VALIDSIG"
    ASSERT_SIGNER = "ASSERT_SIGNER"
    ASSERT_PUBKEY_ALGO = "ASSERT_PUBKEY_ALGO"
    SIG_ID = "SIG_ID"
    ENC_TO = "ENC_TO"
    BEGIN_DECRYPTION = "BEGIN_DECRYPTION"
    END_DECRYPTION = "END_DECRYPTION"
    DECRYPTION_KEY = "DECRYPTION_KEY"
    DECRYPTION_INFO = "DECRYPTION_INFO"
    DECRYPTION_FAILED = "DECRYPTION_FAILED"
    DECRYPTION_OKAY = "DECRYPTION_OKAY"
    SESSION_KEY = "SESSION_KEY"
    BEGIN_ENCRYPTION = "BEGIN_ENCRYPTION"
    END_ENCRYPTION = "END_ENCRYPTION"
    FILE_START = "FILE_START"
    FILE_DONE = "FILE_DONE"
    BEGIN_SIGNING = "BEGIN_SIGNING"
    ALREADY_SIGNED = "ALREADY_SIGNED"
    SIG_CREATED = "SIG_CREATED"
    NOTATION_ = "NOTATION_"
    POLICY_URL = "POLICY_URL"
    PLAINTEXT = "PLAINTEXT"
    PLAINTEXT_LENGTH = "PLAINTEXT_LENGTH"
    ATTRIBUTE = "ATTRIBUTE"
    SIG_SUBPACKET = "SIG_SUBPACKET"
    ENCRYPTION_COMPLIANCE_MODE = "ENCRYPTION_COMPLIANCE_MODE"
    DECRYPTION_COMPLIANCE_MODE = "DECRYPTION_COMPLIANCE_MODE"
    VERIFICATION_COMPLIANCE_MODE = "VERIFICATION_COMPLIANCE_MODE"
    INV_RECP = "INV_RECP"
    INV_SGNR = "INV_SGNR"
    NO_RECP = "NO_RECP"
    NO_SGNR = "NO_SGNR"
    KEY_CONSIDERED = "KEY_CONSIDERED"
    KEYEXPIRED = "KEYEXPIRED"
    KEYREVOKED = "KEYREVOKED"
    NO_PUBKEY = "NO_PUBKEY"
    NO_SECKEY = "NO_SECKEY"
    KEY_CREATED = "KEY_CREATED"
    KEY_NOT_CREATED = "KEY_NOT_CREATED"
    TRUST_ = "TRUST_"
    TOFU_USER = "TOFU_USER"
    TOFU_STATS = "TOFU_STATS"
    TOFU_STATS_SHORT = "TOFU_STATS_SHORT"
    TOFU_STATS_LONG = "TOFU_STATS_LONG"
    PKA_TRUST_GOOD = "PKA_TRUST_GOOD"
    PKA_TRUST_BAD = "PKA_TRUST_BAD"
    GET_BOOL = "GET_BOOL"
    GET_LINE = "GET_LINE"
    GET_HIDDEN = "GET_HIDDEN"
    GOT_IT = "GOT_IT"
    USERID_HINT = "USERID_HINT"
    NEED_PASSPHRASE = "NEED_PASSPHRASE"
    NEED_PASSPHRASE_SYM = "NEED_PASSPHRASE_SYM"
    NEED_PASSPHRASE_PIN = "NEED_PASSPHRASE_PIN"
    MISSING_PASSPHRASE = "MISSING_PASSPHRASE"
    BAD_PASSPHRASE = "BAD_PASSPHRASE"
    GOOD_PASSPHRASE = "GOOD_PASSPHRASE"
    IMPORT_CHECK = "IMPORT_CHECK"
    IMPORTED = "IMPORTED"
    IMPORT_OK = "IMPORT_OK"
    IMPORT_PROBLEM = "IMPORT_PROBLEM"
    IMPORT_RES = "IMPORT_RES"
    EXPORTED = "EXPORTED"
    EXPORT_RES = "EXPORT_RES"
    CARDCTRL = "CARDCTRL"
    SC_OP_FAILURE = "SC_OP_FAILURE"
    SC_OP_SUCCESS = "SC_OP_SUCCESS"
    NODATA = "NODATA"
    UNEXPECTED = "UNEXPECTED"
    TRUNCATED = "TRUNCATED"
    ERROR = "ERROR"
    WARNING = "WARNING"
    NOTE = "NOTE"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    BADARMOR = "BADARMOR"
    DELETE_PROBLEM = "DELETE_PROBLEM"
    PROGRESS = "PROGRESS"
    BACKUP_KEY_CREATED = "BACKUP_KEY_CREATED"
    MOUNTPOINT = "MOUNTPOINT"
    PINENTRY_LAUNCHED = "PINENTRY_LAUNCHED"
    GPGTAR_EXTRACT = "GPGTAR_EXTRACT"
    SIGEXPIRED = "SIGEXPIRED"
    PUBKEY_INFO = "PUBKEY_INFO"
    KEYPAIRINFO = "KEYPAIRINFO"
    CERTINFO = "CERTINFO"
    MANUFACTURER = "MANUFACTURER"
    KEY_STATUS = "KEY-STATUS"
    KEY_ATTR_INFO = "KEY-ATTR-INFO"
    KEY_TIME = "KEY-TIME"
    KEY_LABEL = "KEY-LABEL"


class SigningModes(StrEnum):
    EXPORTABLE = "sign"
    NON_EXPORTABLE = "lsign"
    NON_REVOCABLE = "nrsign"
    TRUST = "tsign"


class RevocationReason(StrEnum):
    NO_REASON = "0"
    USER_ID_INVALID = "4"


class KeyRevocationReason(StrEnum):
    NO_REASON = "0"
    KEY_COMPROMISED = "1"
    KEY_SUPERSEDED = "2"
    KEY_UNUSED = "3"


class KeyTrust(StrEnum):
    UNKNOWN = "1"
    UNTRUSTED = "2"
    MARGINAL_TRUST = "3"
    FULL_TRUST = "4"
    ULTIMATE_TRUST = "5"



class ExecutionError(Exception):
    def __init__(self, output: str, *args: object) -> None:
        self.output = output
        super().__init__(*args)
        
    def __str__(self) -> str:
        return f"Encountered an error executing a GPG command:\n\n=====\n{self.output}\n====="


from gpyg import *


def test_instance_creation(homedir):
    instance = GPG(homedir=homedir, kill_existing_agent=True)
    assert instance.session
    assert instance._config == None
    assert instance.config == instance._config


def test_instance_config(homedir):
    instance = GPG(homedir=homedir, kill_existing_agent=True)
    assert instance.config.version
    assert len(instance.config.compression_algorithms) > 0
    assert len(instance.config.digest_algorithms) > 0
    assert len(instance.config.symmetric_algorithms) > 0
    assert len(instance.config.public_key_algorithms) > 0
    assert len(instance.config.ecc_curves) > 0
    assert instance.keys



from typing import Literal
from pydantic import BaseModel, Field, computed_field
from .infolines import *
from datetime import datetime


class KeyModel(BaseModel):
    """The base KeyModel, containing all relevant information about a key.

    Attributes:
        is_subkey (bool): Whether the key is a subkey
        type (public | secret): What kind of key this is
        validity (FieldValidity): The validity status of the key
        length (int): The key's length
        algorithm (int): The ID of the algorithm used to generate the key
        key_id (str): The key's ID
        creation_date (datetime | None): When the key was created
        expiration_date (datetime | None): When the key will expire
        owner_trust (str | None): The trust value of the key owner
        capabilities (list[KeyCapability]): A list of the key's capabilities
        overall_capabilities (list[KeyCapability]): A list of the key's capabilities, including those of its subkeys.
        curve_name (str | None): ECC curve name
        serial_number (str | None): Key S/N
        fingerprint (str | None): Key fingerprint
        keygrip (str | None): Key keygrip
        all_signatures (list[SignatureInfo]): List of all signatures (revocation & non-revocation)
        user_ids (list[UserIDInfo]): List of this key's user IDs
    """

    is_subkey: bool = False
    type: Literal["public", "secret"]
    validity: FieldValidity
    length: int
    algorithm: int
    key_id: str
    creation_date: datetime | None
    expiration_date: datetime | None
    owner_trust: str | None
    capabilities: list[KeyCapability]
    overall_capabilities: list[KeyCapability]
    curve_name: str | None
    serial_number: str | None = None
    fingerprint: str | None = None
    keygrip: str | None = None
    all_signatures: list[SignatureInfo] = []
    user_ids: list[UserIDInfo] = []
    internal_subkeys: list["KeyModel"] = Field(exclude=True, default_factory=list)

    @computed_field
    @property
    def subkeys(self) -> list["KeyModel"] | None:
        """A list of all subkeys, or None if the key is a subkey.

        Returns:
            list[KeyModel] | None: List of subkeys
        """
        if self.is_subkey:
            return None
        return self.internal_subkeys

    @computed_field
    @property
    def signatures(self) -> list[SignatureInfo]:
        """Gets all non-revocation signatures attached to this key

        Returns:
            list[SignatureInfo]: List of signatures
        """
        return [i for i in self.all_signatures if not i.is_revocation]

    @computed_field
    @property
    def revocation_signatures(self) -> list[SignatureInfo]:
        """Gets all revocation signatures attached to this key

        Returns:
            list[SignatureInfo]: List of revocation signatures
        """
        return [i for i in self.all_signatures if i.is_revocation]

    @staticmethod
    def get_subkeys(
        key: "KeyModel", subkey_map: dict[str, list["KeyModel"]]
    ) -> list["KeyModel"]:
        if key.fingerprint and key.fingerprint in subkey_map.keys():
            for subkey in subkey_map[key.fingerprint]:
                subkey.internal_subkeys = KeyModel.get_subkeys(subkey, subkey_map)
                key.internal_subkeys.append(subkey)
            return key.internal_subkeys
        else:
            return []

    @classmethod
    def from_infolines(cls, lines: list[InfoLine]) -> list["KeyModel"]:
        key_mapping: dict[str, list[KeyModel]] = {"root": []}
        context: KeyModel = None
        for line in lines:
            if line.record_type in [
                InfoRecord.PUBLIC_KEY,
                InfoRecord.SECRET_KEY,
                InfoRecord.SUBKEY,
                InfoRecord.SECRET_SUBKEY,
            ]:
                if context:
                    initial_sigs = [
                        i
                        for i in context.signatures
                        if i.creation_date == context.creation_date
                    ]
                    if (
                        len(initial_sigs) == 0
                        or initial_sigs[0].signer_fingerprint == context.fingerprint
                    ):
                        key_mapping["root"].append(context)
                    else:
                        if not initial_sigs[0].signer_fingerprint in key_mapping.keys():
                            key_mapping[initial_sigs[0].signer_fingerprint] = []
                        key_mapping[initial_sigs[0].signer_fingerprint].append(context)

                context = (
                    KeyModel(
                        type=(
                            "public"
                            if line.record_type == InfoRecord.PUBLIC_KEY
                            else "secret"
                        ),
                        **line.model_dump(),
                    )
                    if line.record_type
                    in [InfoRecord.PUBLIC_KEY, InfoRecord.SECRET_KEY]
                    else KeyModel(
                        type=(
                            "public"
                            if line.record_type == InfoRecord.SUBKEY
                            else "secret"
                        ),
                        is_subkey=True,
                        **line.model_dump(),
                    )
                )
            elif (
                line.record_type
                in [InfoRecord.FINGERPRINT, InfoRecord.SHA256_FINGERPRINT]
                and context
            ):
                context.fingerprint = line.fingerprint
            elif line.record_type == InfoRecord.KEYGRIP and context:
                context.keygrip = line.keygrip
            elif line.record_type == InfoRecord.USER_ID and context:
                context.user_ids.append(line)
            elif line.record_type == InfoRecord.SIGNATURE and context:
                context.all_signatures.append(line)
            elif line.record_type == InfoRecord.REVOCATION_SIGNATURE and context:
                context.all_signatures.append(line)

        if context:
            initial_sigs = [
                i
                for i in context.signatures
                if i.creation_date == context.creation_date
            ]
            if (
                len(initial_sigs) == 0
                or initial_sigs[0].signer_fingerprint == context.fingerprint
            ):
                key_mapping["root"].append(context)
            else:
                if not initial_sigs[0].signer_fingerprint in key_mapping.keys():
                    key_mapping[initial_sigs[0].signer_fingerprint] = []
                key_mapping[initial_sigs[0].signer_fingerprint].append(context)

        results = key_mapping["root"][:]
        for key in results:
            key.internal_subkeys = KeyModel.get_subkeys(key, key_mapping)

        return results



from collections.abc import Generator
from io import BytesIO
import re
import shlex
import subprocess
import threading
import time
from traceback import print_exc
from typing import Any, Literal

class Process:

    def __init__(
        self,
        popen: subprocess.Popen,
        command: str | list[str],
        options: dict[str, Any],
        decode_output: bool = True,
    ):
        self.popen = popen
        self.options = options
        self.command: str = shlex.join(command) if type(command) == list else command
        self.output = "" if decode_output else b""
        self.code: int | None = None
        self.decode = decode_output

    @property
    def pid(self) -> int:
        return self.popen.pid

    def poll(self) -> int | None:
        if self.code == None:
            self.code = self.popen.poll()
        return self.code

    def kill(self):
        try:
            if self.poll() == None:
                self.popen.kill()
        except:
            pass

    def write(self, data: bytes):
        if self.poll() == None:
            self.popen.stdin.write(data)
            self.popen.stdin.flush()

    def wait(self, timeout: float | None = None, kill_on_timeout: bool = True) -> int | None:
        if self.code == None:
            try:
                self.output = (
                    self.popen.communicate(timeout=timeout)[0].decode()
                    if self.decode
                    else self.popen.communicate(timeout=timeout)[0]
                )
            except subprocess.TimeoutExpired:
                if kill_on_timeout:
                    self.kill()

            return self.poll()
        else:
            return self.code

    def send_line(self, line: str):
        if self.poll() == None:
            self.write(line.encode().strip() + b"\n")

class ProcessSession:
    def __init__(self, shell: bool | None = None, environment: dict[str, str] | None = None, working_directory: str | None = None, cleanup_mode: Literal["kill", "wait", "ignore"] = "kill") -> None:
        self.default_options = {
            "shell": shell,
            "env": environment,
            "cwd": working_directory
        }
        self.cleanup = cleanup_mode
        self.processes: dict[int, Process] = {}

    def make_kwargs(self, **passed_kwargs: dict[str, Any]) -> dict[str, Any]:
        result = passed_kwargs.copy()
        for k, v in self.default_options.items():
            if (not k in result.keys() and v != None) or (k in result.keys() and result[k] == None):
                result[k] = v

        return result

    def activate(self):
        self.processes = {}
        return self

    def deactivate(self):
        match self.cleanup:
            case "kill":
                for pid, process in list(self.processes.items()):
                    process.kill()
                    del self.processes[pid]

            case "wait":
                for process in self.processes.values():
                    if process.poll() == None:
                        process.popen.communicate()

            case _:
                pass

    def __enter__(self):
        return self.activate()

    def __exit__(self, *args, **kwargs):
        self.deactivate()

    def parse_cmd(self, cmd: str | list[str], shell: bool) -> str | list[str]:
        if type(cmd) == list:
            result = shlex.join(cmd)
        else:
            result = shlex.join(shlex.split(cmd))

        if not shell:
            result = shlex.split(result)

        return result

    def spawn(
        self,
        command: str | list[str],
        shell: bool | None = None,
        environment: dict[str, str] | None = None,
        working_directory: str | None = None,
        decode: bool = True,
    ) -> Process:
        options = self.make_kwargs(shell=shell, env=environment, cwd=working_directory)
        parsed_command = self.parse_cmd(command, shell=bool(options.get("shell", False)))

        popen = subprocess.Popen(parsed_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **options)
        self.processes[popen.pid] = Process(
            popen, parsed_command, options, decode_output=decode
        )
        return self.processes[popen.pid]

    def run(
        self,
        command: str | list[str],
        shell: bool | None = None,
        environment: dict[str, str] | None = None,
        working_directory: str | None = None,
        timeout: int | None = None,
        decode: bool = True,
        input: str | bytes | None = None,
    ) -> Process:
        options = self.make_kwargs(shell=shell, env=environment, cwd=working_directory)
        parsed_command = self.parse_cmd(
            command, shell=bool(options.get("shell", False))
        )

        popen = subprocess.Popen(
            parsed_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            **options,
        )
        self.processes[popen.pid] = Process(
            popen, parsed_command, options, decode_output=decode
        )

        if input:
            self.processes[popen.pid].write(
                input.encode() if type(input) == str else input
            )

        self.processes[popen.pid].wait(timeout=timeout, kill_on_timeout=True)
        return self.processes[popen.pid]

    def __getitem__(self, pid: int) -> Process:
        return self.processes[pid]



from pydantic import BaseModel

class GPGConfig(BaseModel):
    """A class describing the configuration of the underlying GPG instance.

    Attributes:
        version (str): The version of GPG on the host system
        public_key_algorithms (dict[str, int]): A mapping of {algorithm name: algorithm ID} for all PK algorithms supported by the local GPG implementation.
        symmetric_algorithms (dict[str, int]): A mapping of {algorithm name: algorithm ID} for all symmetric algorithms supported by the local GPG implementation.
        digest_algorithms (dict[str, int]): A mapping of {algorithm name: algorithm ID} for all digest algorithms supported by the local GPG implementation.
        compression_algorithms (dict[str, int]): A mapping of {algorithm name: algorithm ID} for all compression algorithms supported by the local GPG implementation.
        ecc_curves (list[str]): A list of all ECC curve names supported by the local GPG implementation
    """

    version: str
    public_key_algorithms: dict[str, int]
    symmetric_algorithms: dict[str, int]
    digest_algorithms: dict[str, int]
    compression_algorithms: dict[str, int]
    ecc_curves: list[str]

    @classmethod
    def from_config_text(cls, data: str) -> "GPGConfig":
        fields = {line.split(":")[1]: line.split(":")[2].split(";") if ";" in line else line.split(":")[2] for line in data.splitlines() if line.startswith("cfg:")}

        return GPGConfig(
            version=fields["version"],
            public_key_algorithms={name: int(id) for name, id in zip(fields["pubkeyname"], fields["pubkey"])},
            symmetric_algorithms={name: int(id) for name, id in zip(fields["ciphername"], fields["cipher"])},
            digest_algorithms={name: int(id) for name, id in zip(fields["digestname"], fields["digest"])},
            compression_algorithms={name: int(id) for name, id in zip(fields["compressname"], fields["compress"])},
            ecc_curves=fields["curve"]
        )



from tempfile import NamedTemporaryFile
from typing import Literal
from .common import BaseOperator
from .keys import Key
from ..util import ExecutionError


class MessageOperator(BaseOperator):
    def encrypt(
        self,
        data: bytes,
        *recipients: Key | str,
        compress: bool = True,
        format: Literal["ascii", "pgp"] = "ascii",
    ) -> bytes:
        """Encrypt a message to at least one recipient

        Args:
            data (bytes): Data to encrypt
            compress (bool, optional): Whether to compress data. Defaults to True.
            format (ascii | pgp, optional): What format to output. Defaults to "ascii".

        Raises:
            ValueError: If no recipients were specified

        Returns:
            bytes: Encrypted data
        """
        if len(recipients) == 0:
            raise ValueError("Must specify at least one recipient")
        parsed_recipients = " ".join(
            [f"-r {r.key_id if isinstance(r, Key) else r}" for r in recipients]
        )
        cmd = (
            "gpg {compress} --batch --encrypt {recipients} {armored} --output -".format(
                compress="-z 0" if not compress else "",
                recipients=parsed_recipients,
                armored="--armor" if format == "ascii" else "",
            )
        )
        result = self.session.run(cmd, decode=False, input=data)
        if result.code == 0:
            return result.output
        raise ExecutionError(f"Failed to encrypt:\n{result.output}")

    def decrypt(
        self, data: bytes, key: Key | None = None, passphrase: str | None = None
    ) -> bytes:
        """Decrypt PGP-encrypted data

        Args:
            data (bytes): Data to decrypt
            key (Key | None, optional): Recipient key. Defaults to None
            passphrase (str | None, optional): Passphrase, if required. Defaults to None.

        Raises:
            ExecutionError: If the operation fails

        Returns:
            bytes: Decrypted data (with header info removed)
        """
        with NamedTemporaryFile() as datafile:
            datafile.write(data)
            datafile.seek(0)
            cmd = f"gpg {'-u ' + key.fingerprint if key else ''} --batch --pinentry-mode loopback --passphrase-fd 0 --output - --decrypt {datafile.name}"
            result = self.session.run(
                cmd, decode=False, input=passphrase + "\n" if passphrase else None
            )
        if result.code == 0:
            return result.output.split(b"\n", maxsplit=2)[-1]
        raise ExecutionError(f"Failed to decrypt:\n{result.output}")

    def encrypt_symmetric(
        self,
        data: bytes,
        passphrase: str,
        algo: str = "AES",
        format: Literal["ascii", "pgp"] = "ascii",
    ) -> bytes:
        """Symmetrically encrypt data with <algo> and <passphrase>

        Args:
            data (bytes): Data to encrypt
            passphrase (str): Passphrase to use
            algo (str, optional): Algorithm selection. Defaults to "AES".
            format (ascii | pgp, optional): Output format. Defaults to "ascii".

        Raises:
            ExecutionError: If operation fails

        Returns:
            bytes: Encrypted data
        """
        with NamedTemporaryFile() as datafile:
            datafile.write(data)
            datafile.seek(0)
            result = self.session.run(
                f"gpg {'--armor' if format == 'ascii' else ''} --cipher-algo {algo} --output - --passphrase-fd 0 --pinentry-mode loopback --symmetric {datafile.name}",
                decode=False,
                input=passphrase,
            )

            if result.code == 0:
                return result.output
            raise ExecutionError(f"Failed to encrypt:\n{result.output}")

    def get_recipients(
        self,
        data: bytes,
        translate: bool = True,
        include: list[Literal["known", "unknown"]] = ["known", "unknown"],
    ) -> list[Key | str]:
        """Gets all recipients associated with an encrypted message

        Args:
            data (bytes): Encrypted message
            translate (bool, optional): Whether to find existing keys
            include (list[known | unknown], optional): Which keys to include (keys that are known vs keys that are not). Defaults to ["known", "unknown"].

        Raises:
            ExecutionError: If operation fails

        Returns:
            list[Key | str]: List of Key objects or, if none match, key IDs
        """
        with NamedTemporaryFile() as datafile:
            datafile.write(data)
            datafile.seek(0)
            cmd = f"gpg -d --list-only -v {datafile.name}"
            result = self.session.run(cmd)
        if result.code == 0:
            key_ids = [
                i.split()[-1] for i in result.output.split("\n") if "public key is" in i
            ]
            if translate:
                keys = []
                for i in key_ids:
                    existing = self.gpg.keys.get_key(i)
                    if i:
                        if "known" in include:
                            keys.append(existing)
                    else:
                        if "unknown" in include:
                            keys.append(i)

                return keys
            else:
                return key_ids
        raise ExecutionError(f"Failed to get recipients:\n{result.output}")

    def sign(
        self,
        data: bytes,
        key: Key,
        mode: Literal["standard", "clear", "detach"] = "standard",
        passphrase: str | None = None,
        format: Literal["ascii", "pgp"] = "ascii",
    ) -> bytes:
        """Signs data with the specified key.

        Args:
            data (bytes): Data to sign
            key (Key): Key to sign with
            mode (standard | clear | detach, optional): What kind of signature to create. Defaults to "standard".
            passphrase (str | None, optional): Key passphrase, if required. Defaults to None.
            format (ascii | pgp, optional): Output format. Defaults to "ascii".

        Raises:
            ExecutionError: If the operation fails

        Returns:
            bytes: Signed data/detached signature
        """
        with NamedTemporaryFile() as datafile:
            datafile.write(data)
            datafile.seek(0)
            cmd = "gpg --default-key {key} --batch --yes --pinentry-mode loopback {format} --passphrase-fd 0 -o - {operation} {file}".format(
                key=key.key_id,
                format="--armor" if format == "ascii" else "",
                operation={
                    "standard": "--sign",
                    "clear": "--clear-sign",
                    "detach": "--detach-sign",
                }[mode],
                file=datafile.name,
            )
            result = self.session.run(
                cmd, decode=False, input=passphrase + "\n" if passphrase else None
            )
            if result.code == 0:
                return b"\n".join(
                    [
                        i
                        for i in result.output.splitlines()
                        if not i.startswith(b"gpg: ")
                    ]
                )
            raise ExecutionError(f"Failed to sign message:\n{result.output}")

    def verify(
        self,
        data: bytes,
        signature: bytes | None = None,
    ) -> list[tuple[str, str]]:
        """Gets a list of signatures on the given data, with an optional detached signature

        Args:
            data (bytes): Data, or in the case that `signature` is not specified, signed data.
            signature (bytes | None, optional): A detached signature. Defaults to None.

        Returns:
            list[tuple[str, str]]: List of `(Key ID, User ID)` records
        """
        with NamedTemporaryFile() as datafile:
            datafile.write(data)
            datafile.seek(0)
            if signature:
                with NamedTemporaryFile() as sigfile:
                    sigfile.write(signature)
                    sigfile.seek(0)
                    cmd = f"gpg --status-fd 1 --batch -v --verify {sigfile.name} {datafile.name}"
                    result = self.session.run(cmd)
            else:
                cmd = f"gpg --status-fd 1 --batch -v --verify {datafile.name}"
                result = self.session.run(cmd)
        return [
            (i.split(" ")[2], i.split(" ", maxsplit=3)[3])
            for i in result.output.splitlines()
            if i.startswith("[GNUPG:] GOODSIG")
        ]



import pytest
from gpyg import GPG, CardOperator


@pytest.fixture(scope="module")
def scoped_homedir(tmp_path_factory: pytest.TempPathFactory) -> str:
    return str(tmp_path_factory.mktemp("gpg-homedir").absolute())


@pytest.fixture
def homedir(tmp_path_factory: pytest.TempPathFactory) -> str:
    return str(tmp_path_factory.mktemp("gpg-homedir").absolute())


@pytest.fixture(scope="module")
def scoped_instance(scoped_homedir) -> GPG:
    return GPG(homedir=scoped_homedir, kill_existing_agent=True)


@pytest.fixture
def instance(homedir) -> GPG:
    return GPG(homedir=homedir, kill_existing_agent=True)


@pytest.fixture(scope="module")
def environment(scoped_instance: GPG) -> GPG:
    for user in range(4):
        scoped_instance.keys.generate_key(
            name=f"Test User {user}",
            email=f"test-user-{user}@example.com",
            comment=f"Test user # {user}",
            passphrase=f"test-psk-{user}" if user < 2 else None,
        )

    return scoped_instance


@pytest.fixture(scope="module")
def smallenv(scoped_instance: GPG):
    key = scoped_instance.keys.generate_key(
        "user", email="user@example.com", passphrase="user"
    )
    return scoped_instance, key


@pytest.fixture
def interactive(instance: GPG):
    signee = instance.keys.generate_key(
        name="Signee", email="signee@example.com", passphrase="signee"
    )
    signer = instance.keys.generate_key(
        name="Signer", email="signer@example.com", passphrase="signer"
    )
    with signee.edit(user=signer.fingerprint) as editor:
        yield editor, signee, signer



from datetime import datetime
from enum import StrEnum
from itertools import zip_longest
from typing import Any, TypeVar
from typing_extensions import TypedDict
from pydantic import BaseModel, Field, computed_field


class Sex(StrEnum):
    """Describes a SmartCard's sex marker.

    Members:
        - `UNSET`: No sex
        - `FEMALE`: Sex set to female
        - `MALE`: Sex set to male
    """

    UNSET = "u"
    FEMALE = "f"
    MALE = "m"


TPinData = TypeVar("TPinData")


class PinData[TPinData](TypedDict):
    """Generic container for PIN data"""

    pin: TPinData
    reset: TPinData
    admin: TPinData


class UIFData(TypedDict):
    """Smartcard usage data"""

    sign: bool
    decrypt: bool
    auth: bool


class KeyData(TypedDict):
    """Data about an individual key stored on the card"""

    fingerprint: str | None
    created: datetime | None
    keygrip: str | None


class SmartCard(BaseModel):
    """SmartCard Model. Includes parsing of --card-status output."""
    lines: dict[str, list[str] | dict[int, list[str]]]

    def field(
        self, line: str, index: int | None = None, default: Any = None
    ) -> str | None:
        """Gets the value of a field index in a line, with an optional default

        Args:
            line (str): Line key
            index (int | None, optional): Field index, or 0 if None. Defaults to None.
            default (Any, optional): Default to return if not found. Defaults to None.

        Returns:
            str | None: Value at index in field, or None if nonexistent/empty
        """
        try:
            val = self.lines[line][index if index != None else 0]
            if len(val) == 0:
                return default
            return val
        except:
            return default

    @classmethod
    def from_status(cls, status: str) -> "SmartCard":
        """Parses --card-status output and generates a SmartCard

        Args:
            status (str): Status output string

        Returns:
            SmartCard: Generated SmartCard
        """
        lines = {}
        for line in status.splitlines():
            key, *fields = line.split(":")
            if key == "keyattr":
                if not "keyattr" in lines.keys():
                    lines["keyattr"] = {}
                lines["keyattr"][int(fields[0])] = fields[1:]
            else:
                lines[key.lower()] = fields[:]

        return SmartCard(lines=lines)

    @computed_field
    @property
    def reader(self) -> str | None:
        """Reader Name & Info

        Returns:
            str | None: Reader name, or None
        """
        return self.field("reader")

    @computed_field
    @property
    def application_id(self) -> str | None:
        """Application ID

        Returns:
            str | None: Application ID, or None
        """
        return self.field("reader", index=2)

    @computed_field
    @property
    def application_type(self) -> str | None:
        """Application Type

        Returns:
            str | None: Application Type, or None
        """
        return self.field("reader", index=3)

    @computed_field
    @property
    def version(self) -> str | None:
        """Reader Version

        Returns:
            str | None: Reader version, or None
        """
        return self.field("version")

    @computed_field
    @property
    def vendor_id(self) -> str | None:
        """ID of Smartcard Vendor

        Returns:
            str | None: Vendor ID, or None
        """
        return self.field("vendor")

    @computed_field
    @property
    def vendor(self) -> str | None:
        """Name of Smartcard Vendor

        Returns:
            str | None: Vendor name, or None
        """
        return self.field("vendor", index=1)

    @computed_field
    @property
    def serial_number(self) -> str | None:
        """Smartcard S/N

        Returns:
            str | None: Serial number, or None
        """
        return self.field("serial")

    @computed_field
    @property
    def cardholder_name(self) -> str | None:
        """Cardholder Name

        Returns:
            str | None: Cardholder name, or None
        """
        return " ".join(self.lines.get("name")) if "name" in self.lines.keys() else None

    @computed_field
    @property
    def language_preferences(self) -> list[str]:
        """Cardholder Language Preferences

        Returns:
            list[str]: List of language preferences
        """
        value = self.field("lang")
        if value:
            return [
                value[i] + value[i + 1]
                for i in range(0, len(value), 2)
                if len(value) > i + 1
            ]
        else:
            return []

    @computed_field
    @property
    def cardholder_gender(self) -> Sex:
        """Cardholder Sex

        Returns:
            Sex: Specified sex, or UNSET if not specified.
        """
        result = self.field("sex")
        if result:
            return Sex(result)
        return Sex.UNSET

    @computed_field
    @property
    def public_key_url(self) -> str | None:
        """Public Key URI

        Returns:
            str | None: URI, or None
        """
        return self.field("url")

    @computed_field
    @property
    def login_data(self) -> str | None:
        """Login Data

        Returns:
            str | None: Login data, or None
        """
        return self.field("login")

    @computed_field
    @property
    def forced_signature_pin(self) -> bool:
        """Whether signature pin is currently forced

        Returns:
            bool: If pin is forced
        """
        return self.field("forcedpin") == "1"

    @computed_field
    @property
    def key_attrs(self) -> list[tuple[int, int]]:
        """Returns a list of key attrs (ID, Value)

        Returns:
            list[tuple[int, int]]: Attrs (ID, Value)
        """
        attrs = []
        for attr in sorted(list(self.lines.get("keyattr", {}).keys())):
            item = self.lines.get("keyattr", {})[attr]
            attrs.append((int(item[0]), int(item[1])))

        return attrs

    @computed_field
    @property
    def max_pin_lengths(self) -> PinData[int]:
        """Max pin length mapping

        Returns:
            PinData[int]: Mapping of pin type: max length
        """
        return {
            "pin": int(self.field("maxpinlen", index=0, default=0)),
            "reset": int(self.field("maxpinlen", index=1, default=0)),
            "admin": int(self.field("maxpinlen", index=2, default=0)),
        }

    @computed_field
    @property
    def pin_retries(self) -> PinData[int]:
        """Remaining pin retry mapping

        Returns:
            PinData[int]: Mapping of pin type: remaining retries
        """
        return {
            "pin": int(self.field("pinretry", index=0, default=0)),
            "reset": int(self.field("pinretry", index=1, default=0)),
            "admin": int(self.field("pinretry", index=2, default=0)),
        }

    @computed_field
    @property
    def signature_count(self) -> int | None:
        """Count of stored signatures

        Returns:
            int | None: Count, or None
        """
        return int(self.field("sigcount", default=0))

    @computed_field
    @property
    def kdf_setting(self) -> bool:
        """Current KDF setting

        Returns:
            bool: KDF State
        """
        return self.field("kdf") == "on"

    @computed_field
    @property
    def uif_setting(self) -> UIFData:
        """UIF Data (usage)

        Returns:
            UIFData: Mapping of usage: enabled
        """
        return {
            "sign": bool(int(self.field("uif", index=0, default=0))),
            "decrypt": bool(int(self.field("uif", index=1, default=0))),
            "auth": bool(int(self.field("uif", index=2, default=0))),
        }

    @computed_field
    @property
    def stored_keys(self) -> list[KeyData]:
        """Information about the stored keys.

        Returns:
            list[KeyData]: List of KeyData.
        """
        fprs = self.lines.get("fpr", [])
        fprtimes = self.lines.get("fprtime", [])
        grps = self.lines.get("grp", [])

        results: list[KeyData] = []
        for fpr, ftime, grp in zip_longest(fprs, fprtimes, grps):
            if len(fpr) == 0:
                continue
            try:
                created = datetime.fromtimestamp(float(ftime)) if ftime else None
            except:
                created = None
            results.append({"fingerprint": fpr, "created": created, "keygrip": grp})

        return results



import os
from gpyg import *


def test_key_generation(instance):
    result = instance.keys.generate_key(
        "Test User",
        email="test@example.com",
        comment="Test Comment",
        passphrase="test-psk",
    )
    assert result != None
    assert result.type == "public"
    assert len(result.subkeys) == 1


def test_pubkey_list(environment):
    result = environment.keys.list_keys()
    assert len(result) == 4
    assert all([r.type == "public" for r in result])
    assert all([len(r.subkeys) == 1 for r in result])


def test_seckey_list(environment):
    result = environment.keys.list_keys(key_type="secret")
    assert len(result) == 4
    assert all([r.type == "secret" for r in result])
    assert all([len(r.subkeys) == 1 for r in result])


def test_key_reload(environment):
    result = environment.keys.list_keys(key_type="secret")
    for key in result:
        previous = key.fingerprint
        assert previous == key.reload().fingerprint


def test_export(environment):
    result = environment.keys.list_keys()
    exported = result[0].export()
    assert exported.startswith(b"-----BEGIN PGP PUBLIC KEY BLOCK-----")
    assert exported.endswith(b"-----END PGP PUBLIC KEY BLOCK-----")

    result = environment.keys.list_keys(key_type="secret")
    exported = result[0].export(password="test-psk-0")
    assert exported.startswith(b"-----BEGIN PGP PRIVATE KEY BLOCK-----")
    assert exported.endswith(b"-----END PGP PRIVATE KEY BLOCK-----")


def test_expire_key(environment):
    result = environment.keys.list_keys()[0]
    result.set_expiration(expiration=datetime.date(2026, 1, 1), password="test-psk-0")
    assert result.expiration_date.date() == datetime.date(2026, 1, 1)


def test_key_passwords(environment):
    keys = environment.keys.list_keys()
    has_pass = keys[0]
    no_pass = keys[2]
    assert has_pass.is_protected()
    assert has_pass.check_password("test-psk-0")
    assert not has_pass.check_password("test-psk-2")

    assert not no_pass.is_protected()
    assert no_pass.check_password("wrong")


def test_sign_key(environment):
    keys = environment.keys.list_keys()
    keys[0].sign_key(keys[1], password="test-psk-0")
    assert len(keys[1].signatures) == 2


def test_subkeys_wrapped(environment):
    result = environment.keys.list_keys(key_type="secret")
    for key in result:
        assert key.subkeys[0].operator != None


def test_add_uid(environment):
    result = environment.keys.list_keys()[0]
    result.add_user_id(
        uid="Added Test User <atu@example.com> (Test Comment)", passphrase="test-psk-0"
    )
    result.add_user_id(
        name="Second Added Test User", email="satu@example.com", passphrase="test-psk-0"
    )
    assert len(result.user_ids) == 3
    assert "Added Test User <atu@example.com> (Test Comment)" in [
        i.uid for i in result.user_ids
    ]
    assert "Second Added Test User <satu@example.com>" in [
        i.uid for i in result.user_ids
    ]


def test_revoke_uid(environment):
    result = environment.keys.list_keys()[1]
    result.add_user_id(
        uid="Added Test User <atu@example.com> (Test Comment)", passphrase="test-psk-1"
    )
    assert len(result.user_ids) == 2
    result.revoke_uid(
        "Added Test User <atu@example.com> (Test Comment)", passphrase="test-psk-1"
    )
    assert len(result.user_ids) == 2
    assert result.user_ids[1].validity == "r"


def test_set_primary_uid(environment):
    key = environment.keys.list_keys()[3]

    key.add_user_id(uid="New Primary")
    assert len(key.user_ids) == 2
    key.set_primary_uid("New Primary")
    assert key.user_ids[0].uid == "New Primary"


def test_delete_key(environment):
    key = environment.keys.list_keys()[3]
    assert len(key.subkeys) == 1
    key.subkeys[0].delete()
    key.reload()
    assert len(key.subkeys) == 0


def test_import_key(environment):
    keys = environment.keys.list_keys()
    assert len(keys) == 4
    with open(os.path.join(environment.homedir, "export.asc"), "wb") as f:
        f.write(keys[0].export())

    keys[0].delete()
    assert len(environment.keys.list_keys()) == 3
    environment.keys.import_key(os.path.join(environment.homedir, "export.asc"))
    assert len(environment.keys.list_keys()) == 4



def test_instance_working(interactive):
    editor, signee, signer = interactive
    assert editor.key.fingerprint == signee.fingerprint
    assert len(editor.list()) > 0


def test_sign(interactive):
    editor, signee, signer = interactive
    assert len(signee.signatures) == 1
    editor.sign(signer_passphrase="signer")
    editor.save()
    signee.reload()
    assert len(signee.signatures) == 2


def test_delete_sign(interactive):
    editor, signee, signer = interactive
    assert len(signee.signatures) == 1
    editor.sign(signer_passphrase="signer")
    editor.set_uid("1")
    editor.set_key("1")
    editor.delete_signature(signer)
    editor.save()
    signee.reload()
    assert len(signee.signatures) == 1


def test_add_uid(interactive):
    editor, signee, signer = interactive
    assert len(signee.user_ids) == 1
    editor.add_uid("OtherSignee", passphrase="signee")
    editor.save()
    signee.reload()
    assert len(signee.user_ids) == 2


def test_delete_uid(interactive):
    editor, signee, signer = interactive
    assert len(signee.user_ids) == 1
    editor.add_uid("OtherSignee", passphrase="signee")
    editor.set_uid(2)
    editor.delete_uid()
    editor.save()
    signee.reload()
    assert len(signee.user_ids) == 1


def test_delete_key(interactive):
    editor, signee, signer = interactive
    assert len(signee.subkeys) > 0
    editor.set_key("1")
    editor.delete_key()
    editor.save()
    signee.reload()
    assert len(signee.subkeys) == 0



from gpyg import *


def test_encryption(smallenv):
    env, key = smallenv
    DATA = b"test-data"
    encrypted = env.messages.encrypt(b"test-data", key)
    assert encrypted != DATA
    decrypted = env.messages.decrypt(encrypted, key, passphrase="user")
    assert decrypted == DATA


def test_recipients(smallenv):
    env, key = smallenv
    DATA = b"test-data"
    encrypted = env.messages.encrypt(b"test-data", key)
    assert encrypted != DATA

    recipients = env.messages.get_recipients(encrypted)
    assert len(recipients) == 1
    assert isinstance(recipients[0], Key)
    assert recipients[0].key_id == key.key_id


def test_sign(smallenv):
    env, key = smallenv
    DATA = b"test-data"
    encrypted = env.messages.encrypt(b"test-data", key)
    assert encrypted != DATA
    signed = env.messages.sign(encrypted, key, passphrase="user")
    assert signed != encrypted
    verification = env.messages.verify(signed)
    assert len(verification) > 0
    assert verification[0][0] == key.key_id


def test_symmetric(smallenv):
    env, key = smallenv
    DATA = b"test-data"
    encrypted = env.messages.encrypt_symmetric(DATA, "test")
    assert encrypted != DATA
    decrypted = env.messages.decrypt(encrypted, passphrase="test")
    assert decrypted == DATA



from collections.abc import Generator
from contextlib import contextmanager
from datetime import date, datetime, timedelta
from enum import StrEnum
from typing import Any, Literal

from pydantic import Field, PrivateAttr, computed_field

from gpyg.util import interactive
from .common import BaseOperator
from ..util import ExecutionError, ProcessSession, StatusInteractive, StatusLine
from ..models import (
    InfoLine,
    parse_infoline,
    KeyModel,
    StatusCodes,
    SigningModes,
    SignatureInfo,
    RevocationReason,
    KeyRevocationReason,
    KeyTrust,
)


class KeyOperator(BaseOperator):
    def generate_key(
        self,
        name: str,
        email: str | None = None,
        comment: str | None = None,
        algorithm: str | None = None,
        usage: list[Literal["sign", "auth", "encr", "cert"]] | None = None,
        expiration: datetime | timedelta | int | None = None,
        passphrase: str | None = None,
        force: bool = False,
    ) -> "Key | None":
        """Generate a key given a set of parameters.

        Args:
            name (str): UID Name
            email (str | None, optional): Optional UID email. Defaults to None.
            comment (str | None, optional): Optional UID comment. Defaults to None.
            algorithm (str | None, optional): Algorithm name. Defaults to None.
            usage (list["sign" | "auth" | "encr" | "cert"] | None, optional): List of usages, or None for default. Defaults to None.
            expiration (datetime | timedelta | int | None, optional): Key expiration. Defaults to None.
            passphrase (str | None, optional): Key passphrase (if left empty, no passphrase). Defaults to None.
            force (bool, optional): Force creation. Defaults to False.

        Raises:
            ExecutionError: If key generation fails

        Returns:
            Returns the generated key
        """
        uid = "{name}{email}{comment}".format(
            name=name,
            email=f" <{email}> " if email else " ",
            comment=f"({comment})" if comment else "",
        ).strip()

        if isinstance(expiration, datetime):
            expire_str = expiration.isoformat()
        elif isinstance(expiration, timedelta):
            expire_str = "seconds=" + str(expiration.seconds)
        elif type(expiration) == int:
            expire_str = "seconds=" + str(expiration)
        else:
            expire_str = "none"

        command = "gpg {force} --batch --pinentry-mode loopback --passphrase {passphrase} --quick-gen-key '{uid}' {algo} {usage} {expire}".format(
            force="--yes" if force else "",
            passphrase="'" + passphrase + "'" if passphrase else "''",
            uid=uid,
            algo=algorithm if algorithm else "default",
            usage=",".join(usage) if usage else "default",
            expire=expire_str,
        )
        proc = self.session.spawn(command)
        proc.wait()
        if "certificate stored" in proc.output.strip().split("\n")[-1]:
            return self.list_keys(
                pattern=proc.output.strip().split("\n")[-1].split("/")[-1].split(".")[0]
            )[0]
        else:
            raise ExecutionError(proc.output)

    def list_keys(
        self,
        pattern: str = None,
        key_type: Literal["public", "secret"] = "public",
        check_sigs: bool = True,
    ) -> list["Key"]:
        """List keys, optionally filtering by a pattern.

        Args:
            pattern (str | None, optional): Optional pattern to filter results by
            key_type (public | secret, optional): What key type to return
            check_sigs (bool, optional): Whether to check signatures or just list them

        Returns:
            List of results
        """

        args = [
            i
            for i in [
                "gpg",
                "--with-colons",
                "--with-fingerprint",
                "--with-subkey-fingerprint",
                "--with-keygrip",
                "--with-sig-check" if check_sigs else "--with-sig-list",
                f"--list-{key_type}-keys",
                pattern,
            ]
            if i != None
        ]
        proc = self.session.spawn(args)
        proc.wait()

        lines = [i for i in proc.output.splitlines() if not i.startswith("gpg: ")]
        parsed = [parse_infoline(line) for line in lines]
        return Key.from_infolines(self, parsed)

    def get_key(
        self, fingerprint: str, key_type: Literal["public", "secret"] = "public"
    ) -> "Key | None":
        """Gets a specific key given a fingerprint

        Args:
            fingerprint (str): Fingerprint to search for
            key_type (public | secret, optional): What key type to return. Defaults to "public".

        Returns:
            The located Key, or None if not found.
        """
        results = self.list_keys(pattern=fingerprint, key_type=key_type)
        if len(results) == 0:
            return None

        return results[0]

    def import_key(self, *keyfiles: str):
        """Imports keys from file paths into the keyring

        Args:
            *keyfiles (str): Any number of files to import

        Raises:
            ExecutionError: If operation fails
        """
        for file in keyfiles:
            result = self.session.run(f"gpg --batch --yes --import {file}")
            if result.code != 0:
                raise ExecutionError(
                    f"Failed to import {file} with code {result.code}:\n{result.output}"
                )


class Key(KeyModel):
    operator: KeyOperator = Field(exclude=True)
    internal_subkeys: list["Key"] = Field(exclude=True, default_factory=list)
    model_config = {"arbitrary_types_allowed": True}

    @property
    def session(self) -> ProcessSession:
        """Gets the internal ProcessSession instance

        Returns:
            ProcessSession: Internal session
        """
        return self.operator.session

    @computed_field
    def subkeys(self) -> list["Key"] | None:
        if self.is_subkey:
            return None
        return self.internal_subkeys

    @staticmethod
    def apply(operator: KeyOperator, model: KeyModel) -> "Key":
        model.internal_subkeys = [
            Key.apply(operator, i) for i in model.internal_subkeys
        ][:]
        return Key(operator=operator, **dict(model))

    @classmethod
    def from_infolines(
        cls, operator: KeyOperator, lines: list[InfoLine]
    ) -> list["Key"]:
        return [Key.apply(operator, i) for i in super().from_infolines(lines)]

    def reload(self) -> "Key":
        """Reloads cached information from the keyring.

        Raises:
            RuntimeError: Fails if the key has been deleted or is otherwise inaccessible

        Returns:
            Key: A reference to the Key instance
        """
        result = self.operator.get_key(self.fingerprint, key_type=self.type)
        if result:
            for name, value in result.__dict__.items():
                setattr(self, name, value)
        else:
            raise RuntimeError(
                "This key has been deleted/is no longer able to be referenced."
            )

        return self

    def export(
        self,
        mode: Literal["gpg", "ascii"] = "ascii",
        password: str | None = None,
        filters: list[tuple[Literal["keep-uid", "drop-subkey"], str]] = [],
        export_local_sigs: bool | None = None,
        export_attributes: bool | None = None,
        export_sensitive_revkeys: bool | None = None,
        export_backup: bool | None = None,
        export_clean: bool | None = None,
        export_minimal: bool | None = None,
        export_revocs: bool | None = None,
        export_dane: bool | None = None,
        mode1003: bool | None = None,
    ) -> bytes:
        """Exports a key in the given format.

        Args:
            mode (gpg | ascii, optional): Whether to output as GPG binary or ASCII armor. Defaults to "ascii".
            password (str | None, optional): Password required to unlock secret key, which may or may not be relevant. Defaults to None.
            filters (list[tuple[keep-uid | drop-subkey, str]], optional): List of export filters. For more information, see the GPG manual. Defaults to [].

            export_local_sigs (bool | None, optional): Allow exporting key signatures marked as "local". This is not generally useful unless a shared keyring scheme is being used. Defaults to no. Defaults to None.
            export_attributes (bool | None, optional): Include attribute user IDs (photo IDs) while exporting. Not including attribute user IDs is useful to export keys that are going to be used by an OpenPGP program that does not accept attribute user IDs. Defaults to yes. Defaults to None.
            export_sensitive_revkeys (bool | None, optional): Include designated revoker information that was marked as "sensitive". Defaults to no. Defaults to None.
            export_backup (bool | None, optional): Export for use as a backup. The exported data includes all data which is needed to restore the key or keys later with GnuPG. The format is basically the OpenPGP format but enhanced with GnuPG specific data. All other contradicting options are overridden. Defaults to None.
            export_clean (bool | None, optional): Compact (remove all signatures from) user IDs on the key being exported if the user IDs are not usable. Also, do not export any signatures that are not usable. This includes signatures that were issued by keys that are not present on the keyring. This option is the same as running the --edit-key command "clean" before export except that the local copy of the key is not modified. Defaults to no. Defaults to None.
            export_minimal (bool | None, optional): Export the smallest key possible. This removes all signatures except the most recent self-signature on each user ID. This option is the same as running the --edit-key command "minimize" before export except that the local copy of the key is not modified. Defaults to no. Defaults to None.
            export_revocs (bool | None, optional): Export only standalone revocation certificates of the key. This option does not export revocations of 3rd party certificate revocations. Defaults to None.
            export_dane (bool | None, optional): Instead of outputting the key material output OpenPGP DANE records suitable to put into DNS zone files. An ORIGIN line is printed before each record to allow diverting the records to the corresponding zone file. Defaults to None.
            mode1003 (bool | None, optional): Enable the use of a new secret key export format. This format avoids the re-encryption as required with the current OpenPGP format and also improves the security of the secret key if it has been protected with a passphrase. Note that an unprotected key is exported as-is and thus not secure; the general rule to convey secret keys in an OpenPGP encrypted file still applies with this mode. Versions of GnuPG before 2.4.0 are not able to import such a secret file. Defaults to None.

        Raises:
            ExecutionError: Raised if the attempted export operation fails.

        Returns:
            bytes: byte-encoded output.
        """
        options = dict(
            export_local_sigs=export_local_sigs,
            export_attributes=export_attributes,
            export_sensitive_revkeys=export_sensitive_revkeys,
            export_backup=export_backup,
            export_clean=export_clean,
            export_minimal=export_minimal,
            export_revocs=export_revocs,
            export_dane=export_dane,
            mode1003=mode1003,
        )

        cmd = "gpg{format} --pinentry-mode loopback --batch{passphrase} --export-options '{options}' {filters} --export{secret} {fingerprint}".format(
            format=" --armor" if mode == "ascii" else "",
            secret="-secret-keys" if self.type == "secret" else "",
            fingerprint=self.fingerprint,
            passphrase=" --passphrase-fd 0" if password else "",
            options=",".join(
                [
                    ("" if v else "no-") + k.replace("_", "-")
                    for k, v in options.items()
                    if v != None
                ]
            ),
            filters=(
                " ".join(
                    ["--export-filter " + name + "=" + expr for name, expr in filters]
                )
                if len(filters) > 0
                else ""
            ),
        )

        result = self.session.run(
            cmd, decode=False, input=password + "\n" if password else None
        )
        if result.code == 0:
            return result.output.strip()
        raise ExecutionError(result.output.decode())

    def set_expiration(
        self,
        expiration: date | None = None,
        subkeys: list[str] | Literal["*"] | None = None,
        password: str | None = None,
    ) -> "Key":
        """Sets the expiration date of the current Key

        Args:
            expiration (date | None, optional): Expiration date, or None to remove expiration. Defaults to None.
            subkeys (list[str] | Literal[, optional): Which subkeys to apply to. Can be a list of fingerprints, "*", or None for just the primary key. Defaults to None.
            password (str | None, optional): Key password. Defaults to None.

        Raises:
            ExecutionError: Raised if the operation fails

        Returns:
            Key: Updated reference to self
        """
        cmd = "gpg --batch --pinentry-mode loopback --passphrase-fd 0 --quick-set-expire {fingerprint} {expiry} {targets}".format(
            fingerprint=self.fingerprint,
            expiry=expiration.isoformat() if expiration else "0",
            targets=(" ".join(subkeys) if subkeys != "*" else "'*'") if subkeys else "",
        )

        result = self.session.run(cmd, input=password + "\n" if password else None)
        if result.code == 0:
            return self.reload()
        raise ExecutionError(result.output)

    def is_protected(self) -> bool:
        """Checks if the current key is password-protected

        Returns:
            bool: True if protected, False otherwise
        """
        proc = self.session.run(
            f"gpg --dry-run --batch --passphrase-fd 0 --pinentry-mode loopback --passwd '{self.fingerprint}'",
            input="\n",
        )
        return "error" in proc.output

    def check_password(self, password: str) -> bool:
        """Checks whether the given password is valid for this key

        Args:
            password (str): Password to try

        Returns:
            bool: True if correct, False otherwise
        """
        if not self.is_protected():
            return True

        proc = self.session.run(
            f"gpg --dry-run --batch --passphrase-fd 0 --pinentry-mode loopback --passwd '{self.fingerprint}'",
            input=password + "\n",
        )
        return not "error" in proc.output

    def sign_key(
        self,
        target: "str | Key",
        users: list[str] | None = None,
        password: str | None = None,
        exportable: bool = True,
        force: bool = False,
    ) -> "Key":
        """Signs another key using the current key

        Args:
            target (str | Key): Key to sign
            users (list[str] | None, optional): Which users to sign. Defaults to None.
            password (str | None, optional): Password of current key, if needed. Defaults to None.
            exportable (bool, optional): Whether the signature should be exportable. Defaults to True.
            force (bool, optional): Whether to force signing. Defaults to False.

        Raises:
            ExecutionError: Raised if the operation fails.

        Returns:
            Key: Reference to the targeted Key
        """
        if isinstance(target, Key):
            parsed_target = target.fingerprint
        else:
            parsed_target = target

        cmd = "gpg --batch --pinentry-mode loopback --passphrase-fd 0 -u {current} {force} --quick-{local}sign-key {fingerprint} {names}".format(
            current=self.fingerprint,
            force="--force-sign-key" if force else "",
            local="" if exportable else "l",
            fingerprint=parsed_target,
            names=" ".join(['"' + name + '"' for name in users]) if users else "",
        )
        proc = self.session.run(cmd, input=password + "\n" if password else None)
        if proc.code == 0:
            if isinstance(target, Key):
                return target.reload()
            else:
                return self.operator.get_key(target)
        else:
            raise ExecutionError(proc.output)

    def add_subkey(
        self,
        password: str | None = None,
        algorithm: str | None = None,
        usage: list[Literal["sign", "auth", "encr"]] | None = None,
        expiration: datetime | timedelta | int | str | None = None,
        key_passphrase: str | None = None,
    ) -> "Key":
        """Adds a subkey to the selected Key, and refreshes cached data.

        Args:
            password (str | None, optional): Primary key passphrase. Defaults to None.
            algorithm (str | None, optional): Algorithm to use. Defaults to None.
            usage (list[sign | auth | encr] | None, optional): List of usages, or None for the default. Defaults to None.
            expiration (datetime | timedelta | int | str | None, optional): Expiration date, or None for no expiration. Defaults to None.
            key_passphrase (str | None, optional): Passphrase for the new key. Defaults to None.

        Returns:
            Key: Updated reference to self
        """
        if self.is_subkey:
            raise ValueError("Cannot add a subkey to a subkey.")

        if isinstance(expiration, datetime):
            expire_str = expiration.isoformat()
        elif isinstance(expiration, timedelta):
            expire_str = "seconds=" + str(expiration.seconds)
        elif type(expiration) == int:
            expire_str = "seconds=" + str(expiration)
        elif type(expiration) == str:
            expire_str = expiration
        else:
            expire_str = "none"
        cmd = "gpg --batch --pinentry-mode loopback --passphrase-fd 0 --yes --quick-add-key {fingerprint} {algo} {usage} {expiration}".format(
            fingerprint=self.fingerprint,
            algo=algorithm if algorithm else "default",
            usage=",".join(usage) if usage else "default",
            expiration=expire_str,
        )

        proc = self.session.run(
            cmd,
            input="\n".join(
                [password if password else "", key_passphrase if key_passphrase else ""]
            )
            + "\n",
        )
        if proc.code == 0:
            return self.reload()
        else:
            raise ExecutionError(proc.output)

    def add_user_id(
        self,
        uid: str = None,
        name: str = None,
        email: str = None,
        comment: str = None,
        passphrase: str = None,
    ) -> "Key":
        """Add a user ID to the current Key

        Args:
            uid (str, optional): Full UID string (not checked for validity). Defaults to None.
            name (str, optional): Name part. Defaults to None.
            email (str, optional): Email part. Defaults to None.
            comment (str, optional): Comment part. Defaults to None.
            passphrase (str, optional): Key passphrase. Defaults to None.

        Raises:
            ValueError: If both UID and any parts are specified
            ValueError: If neither UID nor Name is specified
            ExecutionError: If the operation fails

        Returns:
            Key: An updated reference to the Key
        """
        if uid and any([name, email, comment]):
            raise ValueError("Cannot specify a full UID and parts at the same time.")
        if not uid and not name:
            raise ValueError("Must at least specify UID or Name.")

        if uid:
            parsed = uid
        else:
            parsed = " ".join(
                [
                    i
                    for i in [
                        name,
                        f"<{email}>" if email else None,
                        f"({comment})" if comment else None,
                    ]
                    if i != None
                ]
            )

        cmd = f"gpg --batch --pinentry-mode loopback --passphrase-fd 0 --quick-add-uid {self.fingerprint} '{parsed}'"
        proc = self.session.run(
            cmd,
            input=passphrase + "\n" if passphrase else None,
        )
        if proc.code == 0:
            return self.reload()
        else:
            raise ExecutionError(proc.output)

    def revoke_uid(self, uid: str, passphrase: str = None) -> "Key":
        """Revokes a given UID on the current Key

        Args:
            uid (str): UID to revoke (full string)
            passphrase (str, optional): Key passphrase, if needed. Defaults to None.

        Raises:
            ExecutionError: If command execution fails

        Returns:
            Key: Reference to updated key
        """
        cmd = f"gpg --batch --pinentry-mode loopback --passphrase-fd 0 --quick-revoke-uid {self.fingerprint} '{uid}'"
        proc = self.session.run(
            cmd,
            input=passphrase + "\n" if passphrase else None,
        )

        if proc.code == 0:
            return self.reload()
        else:
            raise ExecutionError(proc.output)

    def delete(self, delete_both: bool = True) -> None:
        """Deletes self.

        Args:
            delete_both (bool, optional): Whether to delete both the public and private key. Defaults to True.

        Raises:
            ExecutionError: If the operation fails
        """
        cmd = f"gpg --batch --yes {'--delete-secret-and-public-key' if delete_both else ('--delete-keys' if self.type == 'public' else '--delete-secret-keys')} {self.fingerprint}{'!' if self.is_subkey else ''}"
        proc = self.session.run(cmd)

        if proc.code == 0:
            return
        else:
            raise ExecutionError(proc.output)

    def revoke_signature(
        self,
        signer: "Key | str",
        users: list[str] | None = None,
        passphrase: str = None,
    ) -> "Key":
        """Revokes a signature on the current Key generated by Signer

        Args:
            signer (Key | str): The Key (or its fingerprint) that created the signature
            users (list[str] | None, optional): List of users to apply to. If None, applies to all. Defaults to None.
            passphrase (str, optional): Key passphrase, if needed. Defaults to None.

        Raises:
            ExecutionError: If command execution fails

        Returns:
            Key: Reference to updated key
        """
        cmd = "gpg --batch --pinentry-mode loopback --passphrase-fd 0 --quick-revoke-sig {fingerprint} {signer} {names}".format(
            fingerprint=self.fingerprint,
            signer=signer.fingerprint if isinstance(signer, Key) else signer,
            names=" ".join(users) if users else "",
        ).strip()
        proc = self.session.run(
            cmd,
            input=passphrase + "\n" if passphrase else None,
        )
        if proc.code == 0:
            return self.reload()
        else:
            raise ExecutionError(proc.output)

    def set_primary_uid(self, uid: str, passphrase: str | None = None) -> "Key":
        """Set the primary UID of the current Key

        Args:
            uid (str): User ID to set as primary
            passphrase (str | None, optional): Key passphrase, if required. Defaults to None.

        Raises:
            ExecutionError: If command execution fails

        Returns:
            Key: An updated reference to the Key
        """
        cmd = "gpg --batch --pinentry-mode loopback --passphrase-fd 0 --quick-set-primary-uid '{fingerprint}' '{uid}'".format(
            fingerprint=self.fingerprint, uid=uid
        ).strip()
        proc = self.session.run(
            cmd,
            input=passphrase + "\n" if passphrase else None,
        )
        if proc.code == 0:
            return self.reload()
        else:
            raise ExecutionError(proc.output)

    @contextmanager
    def edit(self, user: str | None = None) -> Generator["KeyEditor", Any, Any]:
        """Create an interactive editing instance

        Args:
            user (str | None, optional): Optional user to run as. If left as None, uses the current Key's user ID. Defaults to None.

        Yields:
            KeyEditor: An initialized interactive instance
        """
        with StatusInteractive(
            self.session,
            f"gpg --command-fd 0 --status-fd 1 -u {user if user else self.fingerprint} --with-sig-list --pinentry-mode loopback --yes --with-colons --no-tty --edit-key {self.fingerprint}",
        ) as interactive:
            editor = KeyEditor(self, user if user else self.fingerprint, interactive)
            yield editor


class KeyEditor:

    def __init__(
        self,
        key: Key,
        user: str,
        interactive: StatusInteractive,
    ):
        self.key = key
        self.user = user
        self.interactive = interactive
        self.wait_for_status(StatusCodes.GET_LINE)

    def dbg(self):
        for line in self.interactive.readlines(yield_empty=False):
            print(line)

    def wait_for_status(self, *code: str):
        lines: list[StatusLine] = []
        for line in self.interactive.readlines():
            if line:
                lines.append(line)
                if line.is_status and (len(code) == 0 or line.code in code):
                    return lines

    def list(self) -> list[InfoLine]:
        self.interactive.writelines("list")
        lines = self.wait_for_status(StatusCodes.GET_LINE)
        non_status = [line for line in lines if not line.is_status]
        return [parse_infoline(line.content) for line in non_status]

    def quit(self):
        """Quit without saving"""
        self.interactive.writelines("quit")
        self.interactive.process.wait()

    def save(self):
        """Save changes & quit"""
        self.interactive.writelines("save")
        self.interactive.process.wait()

    def set_uid(self, uid: str):
        """Select a specific UID within the current key

        Args:
            uid (str): UID to select, or "0" for no UID
        """
        self.interactive.writelines(f"uid {uid}")
        self.wait_for_status(StatusCodes.GET_LINE)

    def set_key(self, key: str | Key | None):
        """Select a specific subkey

        Args:
            key (str | Key | None): KEY to select, or None for no selection.
        """

        if isinstance(key, Key):
            self.interactive.writelines(f"key {key.key_id}")
        elif type(key) == str:
            self.interactive.writelines(f"key {key}")
        else:
            self.interactive.writelines("key 0")

        self.wait_for_status(StatusCodes.GET_LINE)

    def sign(
        self,
        mode: SigningModes = SigningModes.EXPORTABLE,
        signer_passphrase: str | None = None,
    ) -> None:
        """Sign the active key.

        Args:
            mode (SigningModes, optional): What mode to use. Defaults to SigningModes.EXPORTABLE.
            signer_passphrase (str | None, optional): Signer passphrase, if applicable. Signer is set with .set_uid(...) Defaults to None.

        Raises:
            ExecutionError: Raised if the operation fails
        """
        self.interactive.writelines(
            str(mode if mode in SigningModes else SigningModes.EXPORTABLE)
        )
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            self.interactive.seek(0)
            raise ExecutionError(output=self.interactive.read().decode())
        self.interactive.writelines("y")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            self.interactive.seek(0)
            raise ExecutionError(output=self.interactive.read().decode())

        if signer_passphrase:
            self.interactive.writelines(signer_passphrase)
        else:
            self.interactive.writelines("")

        self.wait_for_status(StatusCodes.GET_LINE)

    def delete_signature(self, *signers: Key | str):
        """Deletes signatures on the active UID signed by *signers. Does not revoke, just deletes.

        Args:
            *signers (Key | str): Any number of signers (removes all if not specified)

        Raises:
            ExecutionError: If the command fails
        """
        self.interactive.writelines("delsig")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            self.interactive.seek(0)
            raise ExecutionError(
                output="Failed: UID/KEY not selected?\n"
                + self.interactive.read().decode()
            )

        while True:
            if lines[-2].content.startswith("sig:"):
                signature: SignatureInfo = parse_infoline(lines[-2].content)
                if len(signers) == 0 or signature.key_id in [
                    i.key_id if isinstance(i, Key) else i for i in signers
                ]:
                    self.interactive.writelines("y")
                else:
                    self.interactive.writelines("n")
            lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
            if lines[-1].code == StatusCodes.GET_LINE:
                return

    def revoke_signature(
        self,
        *signers: Key | str,
        reason: RevocationReason = RevocationReason.NO_REASON,
        description: str | None = None,
    ):
        """Revokes signatures on the selected UID

        Args:
            *signers (Key | str): List of signers (removes all if not specified)
            reason (RevocationReason, optional): Revocation reason. Defaults to RevocationReason.NO_REASON.
            description (str | None, optional): Revocation description. Defaults to None.

        Raises:
            ExecutionError: If the command fails
        """
        self.interactive.writelines("revsig")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            self.interactive.seek(0)
            raise ExecutionError(
                output="Failed: UID/KEY not selected?\n"
                + self.interactive.read().decode()
            )

        signer_ids = [s.key_id if isinstance(s, Key) else s for s in signers]

        while True:
            request = lines[-1]
            if request.is_status:
                if request.code == StatusCodes.GET_BOOL:
                    match request.arguments[0]:
                        case "ask_revoke_sig.one":
                            sig: SignatureInfo = parse_infoline(lines[-2].content)
                            if len(signers) == 0 or sig.key_id in signer_ids:
                                self.interactive.writelines("y")
                            else:
                                self.interactive.writelines("n")
                        case "ask_revoke_sig.okay":
                            self.interactive.writelines("y")
                        case "ask_revocation_reason.okay":
                            self.interactive.writelines("y")
                        case _:
                            self.interactive.writelines("")
                elif request.code == StatusCodes.GET_LINE:
                    match request.arguments[0]:
                        case "ask_revocation_reason.code":
                            self.interactive.writelines(str(reason))
                        case "ask_revocation_reason.text":
                            if description:
                                self.interactive.writelines(*description.split("\n"))
                            else:
                                self.interactive.writelines("")
                        case "keyedit.prompt":
                            return
            lines = self.wait_for_status()

    def add_uid(
        self,
        real_name: str,
        email: str | None = None,
        comment: str | None = None,
        passphrase: str | None = None,
    ):
        """Adds a single UID to the selected key

        Args:
            real_name (str): UID name
            email (str | None, optional): UID email. Defaults to None.
            comment (str | None, optional): UID comment. Defaults to None.
            passphrase (str | None, optional): Key passphrase, if required. Defaults to None.

        Raises:
            ExecutionError: If the operation fails
        """
        self.interactive.writelines("adduid")
        while True:
            lines = self.wait_for_status()

            if lines[-1].is_status and lines[-1].code == StatusCodes.GET_LINE:
                match lines[-1].arguments[0]:
                    case "keygen.name":
                        self.interactive.writelines(real_name)
                    case "keygen.email":
                        if email:
                            self.interactive.writelines(email)
                        else:
                            self.interactive.writelines("")
                    case "keygen.comment":
                        if comment:
                            self.interactive.writelines(comment)
                        else:
                            self.interactive.writelines("")
                    case "keyedit.prompt":
                        return
            elif lines[-1].is_status and lines[-1].code == StatusCodes.GET_HIDDEN:
                if passphrase:
                    self.interactive.writelines(passphrase)
                else:
                    self.interactive.writelines("")
            elif lines[-1].is_status and lines[-1].code == StatusCodes.ERROR:
                self.interactive.seek(0)
                raise ExecutionError(self.interactive.read().decode())

    def delete_uid(self):
        """Delete the selected UID

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("deluid")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Invalid UID selected")

        self.interactive.writelines("y")
        self.wait_for_status(StatusCodes.GET_LINE)

    def revoke_uid(
        self,
        reason: RevocationReason = RevocationReason.NO_REASON,
        description: str | None = None,
    ):
        """Revokes the selected UID

        Args:
            reason (RevocationReason, optional): Reason for revocation. Defaults to RevocationReason.NO_REASON.
            description (str | None, optional): Optional description. Defaults to None.

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("revuid")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Invalid UID selected")

        self.interactive.writelines("y")
        lines = self.wait_for_status(StatusCodes.GET_LINE)
        self.interactive.writelines(str(reason))

        lines = self.wait_for_status(StatusCodes.GET_LINE)
        self.interactive.writelines(description if description else "")

        lines = self.wait_for_status(StatusCodes.GET_BOOL)
        self.interactive.writelines("y")
        lines = self.wait_for_status(StatusCodes.GET_LINE)

    def set_primary(self):
        """Sets the currently selected UID as primary"""
        self.interactive.writelines("primary")
        self.wait_for_status(StatusCodes.GET_LINE)

    def delete_key(self):
        """Delete the currently selected key

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("delkey")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Key not selected/could not be deleted")
        self.interactive.writelines("y")
        self.wait_for_status(StatusCodes.GET_LINE)

    def revoke_key(
        self,
        reason: KeyRevocationReason = KeyRevocationReason.NO_REASON,
        description: str | None = None,
        passphrase: str | None = None,
    ):
        """Revokes the selected key

        Args:
            reason (KeyRevocationReason, optional): Reason to revoke. Defaults to KeyRevocationReason.NO_REASON.
            description (str | None, optional): Optional description. Defaults to None.
            passphrase (str | None, optional): Key password, if required. Defaults to None.

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("revkey")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Key not selected/could not be deleted")

        self.interactive.writelines("y")
        self.wait_for_status(StatusCodes.GET_LINE)
        self.interactive.writelines(str(reason))
        self.wait_for_status(StatusCodes.GET_LINE)
        self.interactive.writelines(description if description else "")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            return

        self.interactive.writelines(passphrase if passphrase else "")

    def expire_key(self, expires: str = "0", passphrase: str | None = None):
        """Sets a key to expire

        Args:
            expires (str, optional): Amount of time to expire after, or "0" for never. Defaults to "0".
            passphrase (str | None, optional): Passphrase, if required. Defaults to None.

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("expire")
        lines = self.wait_for_status(StatusCodes.GET_LINE)

        self.interactive.writelines(expires)
        self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Invalid value")

        self.interactive.writelines("y")
        self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            return

        if passphrase:
            self.interactive.writelines(passphrase)
        else:
            self.interactive.writelines("")
        self.wait_for_status(StatusCodes.GET_LINE)

    def trust_key(self, amount: KeyTrust = KeyTrust.UNKNOWN):
        """Sets the current key's trust value

        Args:
            amount (KeyTrust, optional): Amount to trust the key. Defaults to KeyTrust.UNKNOWN.

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("trust")
        lines = self.wait_for_status(StatusCodes.GET_LINE)
        if lines[-1].arguments[0] == "keyedit.prompt":
            raise ExecutionError("Unknown error")

        self.interactive.writelines(str(amount))
        lines = self.wait_for_status(StatusCodes.GET_LINE)

    def set_enabled(self, enabled: bool = True):
        """Disable/enable the key

        Args:
            enabled (bool, optional): Whether to enable or disable. Defaults to True.
        """
        self.interactive.writelines("enable" if enabled else "disable")
        self.wait_for_status(StatusCodes.GET_LINE)

    def add_revoker(self, user_id: str):
        """Adds a revoker (cannot be removed)

        Args:
            user_id (str): UID to allow to revoke this key

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("addrevoker")
        lines = self.wait_for_status(StatusCodes.GET_LINE)
        if lines[-1].arguments[0] == "keyedit.prompt":
            raise ExecutionError("Unknown error")

        self.interactive.writelines(user_id)
        self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_BOOL)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Invalid value")

        self.interactive.writelines("y")
        self.wait_for_status(StatusCodes.GET_LINE)

    def change_password(self, old_password: str | None, new_password: str | None):
        """Change selected key's password

        Args:
            old_password (str | None): Old password
            new_password (str | None): New password

        Raises:
            ExecutionError: If operation fails
        """
        self.interactive.writelines("passwd")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Unknown error")

        self.interactive.writelines(old_password if old_password else "")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            raise ExecutionError("Unknown error")

        self.interactive.writelines(new_password if new_password else "")
        lines = self.wait_for_status(StatusCodes.GET_LINE, StatusCodes.GET_HIDDEN)
        if lines[-1].code == StatusCodes.GET_LINE:
            return

        self.interactive.writelines(new_password if new_password else "")
        self.wait_for_status(StatusCodes.GET_LINE)

    def clean(self):
        """Run the `clean` command"""
        self.interactive.writelines("clean")
        self.wait_for_status(StatusCodes.GET_LINE)

    def minimize(self):
        """Run the `minimize` command"""
        self.interactive.writelines("minimize")
        self.wait_for_status(StatusCodes.GET_LINE)



from collections.abc import Generator
from contextlib import contextmanager
import os
import subprocess
from tempfile import TemporaryFile
from .util import *
from .models import *
from .operators import *
from typing import Any

class GPG:
    """Main GPyG class, provides a context within which to perform all operations.

    Args:
        homedir (str | None, optional): Homedir, or the system's default if None. Defaults to None.
        kill_existing_agent (bool, optional): Whether to attempt to kill running GPG agents. Defaults to False.
        write_configs (bool, optional): Whether to write some best-practice configs to the specified homedir (only if a homedir is specified). Defaults to True.
    """

    def __init__(
        self,
        homedir: str | None = None,
        kill_existing_agent: bool = False,
        write_configs: bool = True,
    ) -> None:

        if kill_existing_agent:
            subprocess.run(["gpgconf", "--kill", "gpg-agent"])

        if write_configs and homedir:
            with open(os.path.join(homedir, "scdaemon.conf"), "w") as scdc:
                scdc.write("disable-ccid")
        self.homedir = homedir
        self.session = ProcessSession(environment={"GNUPGHOME": homedir} if homedir else None).activate()
        self._config = None

    @property
    def config(self) -> GPGConfig:
        """Gets the current GPG config

        Returns:
            GPGConfig: Config details
        """
        if not self._config:
            proc = self.session.spawn("gpg --with-colons --list-config")
            proc.wait()
            self._config = GPGConfig.from_config_text(proc.output)
        return self._config

    @property
    def keys(self) -> KeyOperator:
        """Creates a KeyOperator for this instance

        Returns:
            KeyOperator: The KeyOperator
        """
        return KeyOperator(self)

    @property
    def messages(self) -> MessageOperator:
        """Creates a MessageOperator for this instance

        Returns:
            MessageOperator: The MessageOperator
        """
        return MessageOperator(self)

    @contextmanager
    def smart_card(self) -> Generator[CardOperator, Any, None]:
        """Creates a CardOperator with an interactive session attached.

        Yields:
            CardOperator: An initialized CardOperator
        """
        with StatusInteractive(
            self.session,
            "gpg --status-fd 1 --command-fd 0 --pinentry-mode loopback --no-tty --card-edit",
        ) as interactive:
            interactive.wait_for_status(StatusCodes.GET_LINE)
            interactive.writelines("admin")

            yield CardOperator(self, interactive)
            interactive.writelines("quit")



from typing import Any
from ..util import Process, ProcessSession


class BaseOperator:
    def __init__(self, gpg: Any) -> None:
        self.gpg = gpg
        
    @property
    def session(self) -> ProcessSession:
        return self.gpg.session


import os
import shutil
from tempfile import TemporaryDirectory
import time
from gpyg import GPG, StatusInteractive, ProcessSession

if os.path.exists("./tmp"):
    shutil.rmtree("./tmp")

os.makedirs("./tmp", exist_ok=True)
with TemporaryDirectory(dir="tmp", delete=False) as tmpdir:
    gpg = GPG(homedir=tmpdir, kill_existing_agent=True)
    with gpg.smart_card() as card:
        print(card.active)
        if card.active:
            card.reset()
            card.set_usage_info("decrypt", True, "12345678")
            # print(card.active.model_dump_json(indent=4))

    """gpg = GPG(homedir=tmpdir, kill_existing_agent=True)
    key = gpg.keys.generate_key("Bongus", passphrase="test")
    key_other = gpg.keys.generate_key("Bingus", passphrase="test")

    print("BONGUS:", key.fingerprint)
    print("BINGUS:", key_other.fingerprint)
    with key.edit(passphrase="test", run_as=key_other.fingerprint) as editor:
        print(editor.sign())

    key.reload()"""
